from __future__ import annotations
from collections import defaultdict
import csv
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

from attrs import define, asdict, field

if TYPE_CHECKING:
    from .field import Field
    from .parser import Parser

# XX we should convert whole parser.settings to Action subclasses
class Action:
    """ A processing operation. """
    pass

Pivot = str
"Common value in both local and remote columns"

@define
class MergeAction(Action):
    local_column: Field
    rows: Dict[Pivot, List[Union[List[str], Expandable]]]
    remote_parser: Parser

    def get(self, key):
        """ Get the rows containing the key.
        If those do not exist, made up a row full of empty values
        (thanks to the Expandable helper class)
        so that the dimension of the CSV file will not change.
        """
        # Even though rows return list (for an existing key), cast them
        # to list once again. When processing, the duplication mechanism
        # at line `fields[i] *= row_count // len(fields[i])`
        # would duplicate the very list itself (and test_merge would fail)
        # which would result in exponencially bigger line count.
        return list(self.rows.get(key, ())) or [Expandable(("",) * len(self.remote_parser.fields))]

    @classmethod
    def build (cls, remote_file: Path, remote_parser: Parser, remote_column: Field, local_column: Field):
        """ Cache remote values and return a new instance """
        rows = defaultdict(list)
        with remote_file.open() as f:
            reader = csv.reader(f, dialect=remote_parser.dialect)
            for row in reader:
                if not row:  # skip blank
                    continue
                # {'foo': [Expandable(['john@example.com', 'foo']), Expandable(['mary@example.com', 'foo'])],
                # 'bar': [Expandable(['hyacint@example.com', 'bar'])]})
                rows[row[remote_column.col_i]].append(Expandable(row))

        # convert

        return cls(local_column, rows, remote_parser)


class Expandable(list):
    """ A helper class. While merging, multiple columns should be added at once the CSV.
    However, as the pivot value might appear multiple times in the remote file,
    we might want to duplicate the line.
    This class exist to split the columns later, when the line is safely duplicated.
    """
    pass

    @classmethod
    def flatten(cls, xs):
        """ Flatten flattenable elements in an iterable
            https://stackoverflow.com/a/2158532/2036148
            (1, Expandable([2,3])) -> (1, 2, 3)
        """
        for x in xs:
            if isinstance(x, cls):
                yield from cls.flatten(x)
            else:
                yield x


@define
class AggregateAction(Action):
    """
        settings["aggregate"] = column to be grouped, [(sum, column to be summed)]
        Ex: settings["aggregate"] = 1, [(Aggregate.sum, 2), (Aggregate.avg, 3)]
        Ex: settings["aggregate"] = 0, [(Aggregate.count, 0)]
        Ex: settings["aggregate"] = None, [(Aggregate.sum, 1)]
        Ex: settings["aggregate"] = None, [(Aggregate.sum, 1), (Aggregate.avg, 1)]
        """
    group_by : Optional[Field]
    "column to be grouped by"

    actions: List[Tuple[Callable, Field]]
    "[(sum, column to be summed), ...]"


import argparse
from typing import Any, List, Tuple
from sys import exit

from .action_controller import aggregate_functions_str
from .types import Types
from . import __version__

otrs_flags = [("otrs_id", "Ticket id"), ("otrs_num", "Ticket num"), ("otrs_cookie", "OTRSAgentInterface cookie"),
                     ("otrs_token", "OTRS challenge token")]

new_fields: List[Tuple[bool, Any]] = []
"User has requested to compute these. Defined by tuples: add (whether to include the column in the result), field definition"

class BlankTrue(argparse.Action):
    """ When left blank, this flag produces True. (Normal behaviour is to produce None which I use for not being set.)
        Return boolean for 0/false/off/1/true/on.
        Return a metavar value if metavar is a list.
        Else raises ValueError.
    """

    def __call__(self, _, namespace, values, option_string=None, allow_string=False):
        if values in [None, []]:  # blank argument with nargs="?" produces None, with ="*" produces []
            values = True
        elif values.lower() in ["0", "false", "off"]:
            values = False
        elif values.lower() in ["1", "true", "on"]:
            values = True
        elif not allow_string \
                and (type(self.metavar) is not list or values.lower() not in self.metavar) \
                and (len(self.metavar.split("/")) < 2 or values.lower() not in self.metavar.split("/")):
            print(f"Unrecognised value '{values}' of '{self.dest}'. Allowed values are 0/1/BLANK."
                  f" Should the value be considered a positional parameter, move '{self.dest}' behind.")
            exit()
        setattr(namespace, self.dest, values)


class BlankTrueString(BlankTrue):
    """ When left blank, this flag produces True.
        Return boolean for 0/false/off/1/true/on.
        Else returns input value or None if flag omitted.
    """

    def __call__(self, *args, **kwargs):
        super().__call__(*args, **kwargs, allow_string=True)

class FieldExcludedAppend(argparse.Action):
    def __call__(self, _, namespace, values, option_string=None):
        new_fields.append((False, values))


class FieldVisibleAppend(argparse.Action):
    def __call__(self, _, namespace, values, option_string=None):
        new_fields.append((True, values))


class SmartFormatter(argparse.HelpFormatter):

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)



class ArgsController:
    def parse_args(self):
        epilog = "To launch a web service see README.md."
        column_help = "COLUMN is ID of the column (1, 2, 3...), position from the right (-1, ...)," \
                      " the exact column name, field type name or its usual name."
        parser = argparse.ArgumentParser(description="Swiss knife for mutual conversion of the web related data types, like `base64` or outputs of the programs `whois`, `dig`, `curl`. Convenable way to quickly gather all meaningful information or to process large files that might freeze your spreadsheet processor.", formatter_class=SmartFormatter,
                                         epilog=epilog)

        group = parser.add_argument_group("Input/Output")
        group.add_argument('file_or_input', nargs='?', help="File name to be parsed or input text. "
                                                            "In nothing is given, user will input data through stdin.")
        group.add_argument('--file', help="Treat <file_or_input> parameter as a file, never as an input",
                           action="store_true")
        group.add_argument('-i', '--input', help="Treat <file_or_input> parameter as an input text, not a file name",
                           action="store_true")
        group.add_argument('-o', '--output', help="Save output to this file."
                                                  " If left blank, pass output to STDOUT."
                                                  " If omitted, a filename will be produced automatically."
                                                  " May be combined with --headless.",
                           action=BlankTrueString, nargs="?", metavar="blank/FILENAME")
        group.add_argument('-S', '--single-query', help="Consider the input as a single value, not a CSV.",
                           action="store_true")
        group.add_argument('--single-detect', help="Consider the input as a single value, not a CSV,"
                                                   " and just print out possible types of the input.", action="store_true")
        group.add_argument('-C', '--csv-processing', help="Consider the input as a CSV, not a single.",
                           action="store_true")

        group = parser.add_argument_group("CLI experience")
        group.add_argument('--debug', help="Development only: increases verbosity and gets the prompt in the case of an exception.",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--crash-post-mortem', help="Get prompt if program crashes",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('-v', '--verbose', help="Sets the verbosity to see DEBUG messages.", action="store_true")
        group.add_argument('-q', '--quiet', help="R|Sets the verbosity to see WARNINGs and ERRORs only."
                                                 " Prints out the least information possible."
                                                 "\n(Ex: if checking single value outputs a single word, prints out just that.)",
                           action="store_true")
        group.add_argument('-y', '--yes', help="Assume non-interactive mode and the default answer to questions. "
                                               "Will not send e-mails unless --send is on too.",
                           action="store_true")
        group.add_argument('-H', '--headless',
                           help="Launch program in a headless mode which imposes --yes and --quiet. No menu is shown.",
                           action="store_true")
        group.add_argument('--compute-preview', help="When adding new columns, show few first computed values.",
                           action=BlankTrue, nargs="?", metavar="blank/false")

        group = parser.add_argument_group("Environment")
        group.add_argument('--config', help="R|Open a config file and exit."
                                            "\n File: config (default)/uwsgi/template/template_abroad"
                                            "\n Mode: 1 terminal / 2 GUI / 3 try both (default)",
                           nargs='*', metavar=("FILE", "MODE"))
        group.add_argument('--show-uml', help="R|Show UML of fields and methods and exit."
                                              " Methods that are currently disabled via flags or config file are grayed out."
                                              "\n * FLAGs:"
                                              "\n    * +1 to gray out disabled fields/methods"
                                              "\n    * +2 to include usual field names",
                           type=int, const=1, nargs='?')
        group.add_argument('--get-autocompletion', help="Get bash autocompletion.", action="store_true")
        group.add_argument('--version', help=f"Show the version number (which is currently {__version__}).",
                           action="store_true")

        group = parser.add_argument_group("Processing")
        group.add_argument('--threads', help="Set the thread processing number.",
                           action=BlankTrueString, nargs="?", metavar="blank/false/auto/INT")
        group.add_argument('-F', '--fresh', help="Do not attempt to load any previous settings / results."
                                                 " Do not load convey's global WHOIS cache."
                                                 " (But merge WHOIS results in there afterwards.)", action="store_true")
        group.add_argument('-R', '--reprocess', help="Do not attempt to load any previous settings / results."
                                                     " But load convey's global WHOIS cache.", action="store_true")

        parser.add_argument('--server', help=f"Launches simple web server", action="store_true")
        parser.add_argument('--daemon', help=f"R|Run a UNIX socket daemon to speed up single query requests."
                                             "\n  * 1/true/on – allow using the daemon"
                                             "\n  * 0/false/off – do not use the daemon"
                                             "\n  * start – start the daemon and exit"
                                             "\n  * stop – stop the daemon and exit"
                                             "\n  * status – print out the status of the daemon"
                                             "\n  * restart – restart the daemon and continue"
                                             "\n  * server – run the server in current process (I.E. for debugging)",
                            action=BlankTrue, nargs="?", metavar="start/restart/stop/status/server")

        group = parser.add_argument_group("CSV dialect")
        group.add_argument('--delimiter', help="Treat file as having this delimiter. For tab use either \\t or tab.")
        group.add_argument('--quote-char', help="Treat file as having this quoting character")
        group.add_argument('--header', help="Treat file as having header", action="store_true")
        group.add_argument('--no-header', help="Treat file as not having header", action="store_true")
        group.add_argument('--delimiter-output', help="Output delimiter. For tab use either \\t or tab.",
                           metavar="DELIMITER")
        group.add_argument('--quote-char-output', help="Output quoting char", metavar="QUOTE_CHAR")
        group.add_argument('--header-output', help="If false, header is omitted when processing..",
                           action=BlankTrue, nargs="?", metavar="blank/false")

        group = parser.add_argument_group("Actions")
        group.add_argument('-d', '--delete',
                           help="Delete a column. You may comma separate multiple columns."
                           "\n " + column_help, metavar="COLUMN,[COLUMN]")
        group.add_argument('-f', '--field',
                           help="R|Compute field."
                                "\n* FIELD is a field type (see below) that may be appended with a [CUSTOM] in square brackets."
                                "\n* " + column_help +
                                "\n* SOURCE_TYPE is either field type or usual field type. "
                                "That way, you may specify processing method."
                                "\n* CUSTOM is any string dependent on the new FIELD type (if not provided, will be asked it for)."
                                "\nEx: --field tld[gTLD]  # would add TLD from probably a hostname, filtered by CUSTOM=gTLD"
                                "\nEx: --field netname,ip  # would add netname column from any IP column"
                                "\n    (Note the comma without space behind 'netname'.)"
                                "\n\nComputable fields: " + "".join(
                                    "\n* " + t.doc() for t in Types.get_computable_types()) +
                           "\n\nThis flag May be used multiple times.",
                           action=FieldVisibleAppend, metavar="FIELD,[COLUMN],[SOURCE_TYPE],[CUSTOM],[CUSTOM]")
        group.add_argument('-fe', '--field-excluded',
                           help="The same as field but its column will not be added to the output.",
                           action=FieldExcludedAppend, metavar="FIELD,[COLUMN],[SOURCE_TYPE],[CUSTOM],[CUSTOM]")
        group.add_argument('-t', '--type', help="R|Determine column type(s)."
                                                "\nEx: --type country,,phone"
                                                " # 1st column is country, 2nd unspecified, 3rd is phone",
                           metavar="[TYPE],...")
        group.add_argument('--split', help="Split by this COLUMN.",
                           metavar="COLUMN")
        group.add_argument('-s', '--sort', help="List of columns.",
                           metavar="COLUMN,...")
        group.add_argument('-u', '--unique', help="Cast unique filter on this COLUMN.",
                           metavar="COLUMN,VALUE")
        group.add_argument('-ef', '--exclude-filter', help="Filter include this COLUMN by a VALUE.",
                           metavar="COLUMN,VALUE")
        group.add_argument('-if', '--include-filter', help="Filter include this COLUMN by a VALUE.",
                           metavar="COLUMN,VALUE")
        group.add_argument('-a', '--aggregate', help="R|Aggregate"
                                                     "\nEx: --aggregate 2,sum # will sum the second column"
                                                     "\nEx: --aggregate 2,sum,3,avg # will sum the second column and average the third"
                                                     "\nEx: --aggregate 2,sum,1 # will sum the second column grouped by the first"
                                                     "\nEx: --aggregate 1,count # will count the grouped items in the 1st column"
                                                     " (count will automatically set grouping column to the same)"
                                                     f"\n\nAvailable functions: {aggregate_functions_str}",
                           metavar="[COLUMN, FUNCTION], ..., [group-by-COLUMN]")
        group.add_argument('--merge', help="R|Merge another file here. "
                           "\n " + column_help,
                           metavar="[REMOTE_PATH],[REMOTE_COLUMN],[LOCAL_COLUMN]")

        group = parser.add_argument_group("Enabling modules")
        group.add_argument('--whois',
                           help="R|Allowing Whois module: Leave blank for True or put true/on/1 or false/off/0.",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--nmap',
                           help="R|Allowing NMAP module: Leave blank for True or put true/on/1 or false/off/0.",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--dig',
                           help="R|Allowing DNS DIG module: Leave blank for True or put true/on/1 or false/off/0.",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--web', help="R|Allowing Web module: Leave blank for True or put true/on/1 or false/off/0."
                                         "\nWhen single value input contains a web page, we could fetch it and add"
                                         " status (HTTP code) and text fields. Text is just mere text, no tags, style,"
                                         " script, or head. ",
                           action=BlankTrue, nargs="?", metavar="blank/false")

        group = parser.add_argument_group("Field computing options")
        group.add_argument('--disable-external',
                           help="R|Disable external function registered in config.ini to be imported.",
                           action="store_true", default=False)
        group.add_argument('--json', help="When checking single value, prefer JSON output rather than text.",
                           action="store_true")
        group.add_argument('--user-agent', help="Change user agent to be used when scraping a URL")
        group.add_argument('--multiple-hostname-ip', help="Hostname can be resolved into multiple IP addresses."
                                                          " Duplicate row for each.",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--multiple-cidr-ip', help="CIDR can be resolved into multiple IP addresses."
                                                      " Duplicate row for each.",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--web-timeout', help="Timeout used when scraping a URL", type=int, metavar="SECONDS")

        group = parser.add_argument_group("WHOIS module options")
        group.add_argument('--whois-ttl', help="How many seconds will a WHOIS answer cache will be considered fresh.",
                           type=int, metavar="SECONDS")
        group.add_argument('--whois-delete', help="Delete convey's global WHOIS cache.", action="store_true")
        group.add_argument('--whois-delete-unknown', help="Delete unknown prefixes from convey's global WHOIS cache.",
                           action="store_true")
        group.add_argument('--whois-reprocessable-unknown',
                           help="Make unknown lines reprocessable while single file processing,"
                                " do not leave unknown cells empty.", action="store_true")
        group.add_argument('--whois-cache', help="Use whois cache.", action=BlankTrue, nargs="?", metavar="blank/false")

        group = parser.add_argument_group("Sending options")
        group.add_argument('--send', help="Automatically send e-mails when split.",
                           action=BlankTrueString, nargs="?", metavar="blank/smtp/otrs")
        group.add_argument('--send-test', help="Display e-mail message that would be generated for given e-mail.",
                           nargs=2, metavar=("E-MAIL", "TEMPLATE_FILE"))
        group.add_argument('--jinja', help="Process e-mail messages with jinja2 templating system",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--attach-files', help="Split files are added as e-mail attachments",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--attach-paths-from-path-column', help="Files from a column of the Path type are added as e-mail attachments."
                           " Note for security reasons, files must not be symlinks, be readable for others `chmod o+r`, and be in the same directory."
                           " So that a crafted CSV would not pull up ~/.ssh or /etc/ files.",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--testing', help="Do not be afraid, e-mail messages will not be sent."
                                             " They will get forwarded to the testing e-mail"
                                             " (and e-mails in Cc will not be sent at all)",
                           action=BlankTrue, nargs="?", metavar="blank/false")
        group.add_argument('--subject',
                           help="E-mail subject."
                                " May be in BASE64 if started with \"data:text/plain;base64,\"", metavar="SUBJECT")
        group.add_argument('--body',
                           help="E-mail body text (or HTML)."
                                " May be in BASE64 if started with \"data:text/plain;base64,\"", metavar="TEXT")
        group.add_argument('--references',
                           help="E-mail references header (to send message within a thread)."
                                " If used, Bcc header with the `email_from_name` is added to the e-mail template.",
                           metavar="MESSAGE_ID")

        group = parser.add_argument_group("OTRS")
        for flag in otrs_flags:
            group.add_argument('--' + flag[0], help=flag[1])
        return parser



# This file should not do anything that is not vitally needed to stay fast. (Import config would take us 10 ms, do not do it.)
import struct

# XX I suppose this will not work on Win. What other paths should I use instead?
# Importing default package tempdir would cost another 7 ms which is a lot.
socket_file = "/tmp/convey_socket"


def daemon_pid():
    import subprocess
    return subprocess.run(["lsof", "-t", socket_file], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).stdout.strip()


def send(pipe, msg):
    d = msg.encode("utf-8")
    msg = struct.pack('>I', len(d)) + d
    try:
        pipe.sendall(msg)
    except BrokenPipeError:
        return False
    return True


def recv(pipe):
    def recv(n):
        # Helper function to recv n bytes or return None if EOF is hit
        data = b''
        while len(data) < n:
            packet = pipe.recv(n - len(data))
            if not packet:
                return None
            data += packet
        return data

    raw_msglen = recv(4)
    if not raw_msglen:
        pipe.close()
        return False
    return recv(struct.unpack('>I', raw_msglen)[0]).decode("utf-8")



import csv
import logging
import string
from pathlib import Path
from typing import List, Optional
from sys import exit

from dialog import Dialog, DialogError, ExecutableNotFound

from .config import Config, get_terminal_size

logger = logging.getLogger(__name__)

try:
    dialog = Dialog(autowidgetsize=True)
except ExecutableNotFound:
    print("\nError importing dialog library. Try installing: `sudo apt install dialog`.")
    exit()


# skips the dialog in case there is a single value
def skippable_menu(*args, skippable=True, **kwargs):
    """
    :param self: Dialog
    :type skippable: bool If True and there is single option, the dialog returns 'Ok' without asking user.
    """
    if skippable and kwargs["choices"] and len(kwargs["choices"]) == 1:
        return "ok", kwargs["choices"][0][0]
    return dialog.menu(*args, **kwargs)


class Cancelled(Exception):
    pass


class Debugged(Exception):
    pass

def pick_option(options, title="", guesses: Optional[List]=None, skippable=True):
    """ Loop all options
        options tuples of items and descriptions: [("url", "this is url")]
        guesses = indices of options that should be highlighted
        returns option number OR raises Cancelled

        :type skippable: bool If True and there is single option, the dialog returns 'Ok' without asking user.
    """
    if not guesses:
        guesses = []

    # convert numbers `0-8` → `1-9` and `9-...` → `a-...` and then back to number
    abc = string.ascii_lowercase
    i_to_abc = lambda i: abc[i + 1 - 10] if (i + 1 >= 10 and i + 1 - 10 < len(abc)) else i + 1
    abc_to_i = lambda s: int(s) - 1 if s.isdigit() else abc.index(s) + 10 - 1

    choices = []
    if guesses:
        # X SORTED GUESSES:
        opts = [it for it in enumerate(options)]
        for g in guesses:
            if g < len(opts):
                choices.append((f"{i_to_abc(g)} * {opts[g][1][0]} *", str(opts[g][1][1])))

    if not len(guesses) == len(options):  # if every column highlighted, no need to list them all just again
        if guesses:
            title += "\n\nAutomatically detected fields on the top"
            choices.append(("-", "-----"))
        for i, (field_name, desc) in enumerate(options):
            choices.append((f"{i_to_abc(i)} {field_name}", str(desc)))

    code, col_i = skippable_menu(title or " ", choices=choices, skippable=skippable)
    if col_i == '-':
        return pick_option(options, title, guesses, skippable)
    if code != "ok":
        raise Cancelled(".. no column chosen")
    col_i = col_i.split(" ")[0]
    return abc_to_i(col_i)


def ask_number(text):
    """
    Let user write number. Empty input = 0.
    """
    while True:
        try:
            t = ask(text=text)
            if not t:
                return 0
            return int(t)
        except ValueError:
            print("This is not a number")


def ask(text=None):
    try:
        txt = input(text + " ") if text else input()
    except EOFError:
        txt = "x"
    if txt == "x":
        raise Cancelled(".. cancelled")
    if txt == "debug":
        raise Debugged("lets debug")
    return txt


def is_yes(text):
    if Config.get("yes", get=bool):
        return True
    if Config.get("daemon", get=bool):
        raise ConnectionAbortedError
    return ask(text=text + " [y]/n: ").lower() in ("y", "yes", "")


def is_no(text):
    if Config.get("yes", get=bool):
        return True
    if Config.get("daemon", get=bool):
        raise ConnectionAbortedError
    return ask(text=text + " y/[n]: ").lower() in ("n", "no", "")

def hit_any_key(text: str):
    """ Display text and let the user hit any key. Skip when headless. """
    if Config.get("yes", get=bool):
        logger.info(text)
        return
    input(text + " Hit any key.")


class Menu:
    def __init__(self, title=None, callbacks=True, fullscreen=False, skippable=True):
        """ self.menu of tuples [(key, title, f), ("1", "First option", lambda), ...]
        :type skippable: bool If True and there is single option, the dialog returns 'Ok' without asking user.
        """
        self.title = title
        self.menu = []
        self.callbacks = callbacks
        self.skippable = skippable
        self._keyCount = 0
        self.fullscreen = fullscreen
        self.default = None

    def add(self, title, fn=None, key=None, default=False):
        """ Add new item to the menu.

            key - available through this letter
                - if no key set, it will be available through the add order number
                - if False, not available

            default - If user hits enter, the last item with default=True will be triggered.

            """
        if default:
            self.default = key
            title += " ←←←←←"
        if key is None:
            self._keyCount += 1
            key = str(self._keyCount)
        self.menu.append((str(key), title, fn))

    def sout(self, session=None, options={}):
        while True:
            if self.title:
                print("\n" + self.title)
            l = []
            if len(self.menu) == 0:
                input("No possible choices. Continue...")
                return
            for key, name, f in self.menu:
                if key is False or (self.callbacks and not f):
                    l.append(("~", name, False))
                    print("~) {}".format(name))
                    continue
                l.append((key, name, False))
                print("{}) {}".format(key, name))
            try:
                if self.fullscreen:
                    code, ans = skippable_menu(self.title, choices=[(it[0], it[1]) for it in l], skippable=self.skippable)
                    if code != "ok":
                        return
                elif session:
                    ans = session.prompt("? ", **options)
                    if ans == "refresh":
                        return
                else:
                    ans = input("? ")
            except EOFError:
                ans = "x"
            print()
            if ans == "" and self.default:
                ans = self.default
            for key, name, f in self.menu:
                if key == ans:
                    if self.callbacks:
                        if not f:
                            continue
                        f()
                        return True
                    else:
                        return key

            else:
                if ans == "x":
                    return
                elif ans == "debug":
                    Config.get_debugger().set_trace()
                    return
                print("Invalid option")

def choose_file(title:str, dialog: Optional[Dialog]=None):
    if not dialog:
        dialog = Dialog(autowidgetsize=True)
    try:
        code, path = dialog.fselect(str(Path.cwd()), title=title)
    except DialogError as e:
        try:  # I do not know why, fselect stopped working and this helped
            code, path = dialog.fselect(str(Path.cwd()), title=title,
                                                            height=max(get_terminal_size()[0] - 20, 10))
        except DialogError as e:
            input("Unable launch file dialog. Please post an issue to the Github! Hit any key...")
            raise Cancelled("... cancelled")

    if code != "ok" or not path:
        raise Cancelled("... cancelled")
    return path

def csv_split(val):
    """ Sometimes ",".split is not enough, they can use quotes and commas in our mighty CLI. """
    return list(csv.reader([val]))[0]


import logging
import re
import smtplib
import sys
from abc import abstractmethod, ABC
from socket import gaierror
from bs4 import BeautifulSoup
from requests import get, post

from validate_email import validate_email

from envelope import Envelope


from .dialogue import is_yes
from .parser import Parser
from .attachment import Attachment
from .config import Config

re_title = re.compile('<title>([^<]*)</title>')
logger = logging.getLogger(__name__)

OTRS_VERSION = 6


class MailSender(ABC):
    def __init__(self, parser):
        self.parser: Parser = parser

    def start(self):
        pass

    def stop(self):
        pass

    @abstractmethod
    def process(self):
        pass

    def send_list(self, mails):
        """ Send a bunch of e-mail messages
        :type mails: Union[list, generator]
        """
        sent_mails = 0
        total_count = 0
        status = None

        if self.start() is False:
            return False
        for attachment in iter(mails):  # make sure mails are generator to correctly count total_count
            attachment: Attachment

            try:
                e: Envelope = attachment.get_envelope()
            except RuntimeError as e:
                # troubles with an attachment
                total_count += 1
                logging.error('Troubles sending the mail for %s. %s', attachment.mail, e)
                continue

            if e.message().strip() == "" or e.subject().strip() == "":
                # program flow should never allow lead us here
                print("Missing subject or mail body text.")
                return False

            if attachment.sent:
                # program flow should never allow lead us here
                print(f"Message for {attachment.mail} already sent, skipping!")
                continue

            for address in e.to():
                if not validate_email(address, check_dns=False, check_smtp=False):
                    logger.error("Erroneous e-mail: {}".format(address))
                    continue

            status = False
            # noinspection PyBroadException
            try:
                status = self.process(e)
            except KeyboardInterrupt:
                status = "interrupted"
                total_count += len(list(mails))
                break
            except Exception as e:
                logger.error(e)
                Config.error_caught()
                # if something fails (attachment FileNotFoundError, any other error),
                # we want to tag the previously sent e-mails so that they will not be resend next time
            finally:
                if status:
                    t = "Marking as sent" if status == "interrupted" else "Sent"
                    s = ", ".join(e.recipients())
                    if Config.is_testing():
                        s += f" (intended for {attachment.mail})"
                    logger.warning(f"{t}: {s}")
                    sent_mails += 1
                    attachment.sent = True
                else:
                    logger.error(f"Error sending: {attachment.mail}")
                    attachment.sent = False
                total_count += 1
        self.stop()

        if status == "interrupted":
            print(f"Interrupted! We cannot be sure if the e-mail {attachment.mail} was sent but it is probable.")

        print("\nSent: {}/{} mails.".format(sent_mails, total_count))
        if sent_mails != total_count:
            print("Could not send all abroad mails. (Details in convey.log.)")


class MailSenderOtrs(MailSender):

    def _post_multipart(self, url, fields, cookies, attachments):
        """
        Post fields and files to an http host as multipart/form-data.
        fields is a sequence of (name, value) elements for regular form fields.
        files is a sequence of (name, filename, value) elements for data to be uploaded as files.
        Return an appropriate http.client.HTTPResponse object.
        """

        # import ssl
        # # XX once upon a day (10.2.2017), the certificate stopped working or whatever. This is an ugly solution - i think we may delete this line in few weeks
        # If this happens again, we may add a check that we are able to connect to OTRS before sending starts.
        # ssl._create_default_https_context = ssl._create_unverified_context
        # context = ssl._create_unverified_context()
        # context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        # context.options &= ~ssl.OP_NO_TLSv1

        # if Config.is_debug():
        # http.client.HTTPConnection.debuglevel = 1
        # requests_log = logging.getLogger("requests.packages.urllib3")
        # requests_log.setLevel(logging.INFO)
        # requests_log.propagate = True

        # upload attachments before the main request
        if attachments:
            # get FormID to pair attachments
            logger.debug("Receiving FormID")
            upload_form = get(url,
                              params={"ChallengeToken": self.parser.otrs_token,
                                      "Action": "AgentTicketForward",
                                      "TicketID": str(self.parser.otrs_id)},
                              cookies=cookies
                              )
            m = re.search(r'FormID" value="((\d|\.)*)"', upload_form.text)
            if m:
                form_id = m[1]
                self._clean_attachments(upload_form, url, form_id, cookies)
                for a in attachments:
                    logger.debug("Uploading %s", a.name)
                    post(url,
                         params={"Action": "AjaxAttachment",
                                 "Subaction": "Upload",
                                 "FormID": form_id,
                                 "ChallengeToken": self.parser.otrs_token
                                 },
                         cookies=cookies,
                         files={"Files": (a.name, a.data)}
                         )

                fields["FormID"] = form_id
            else:
                raise RuntimeError("Cannot get the FormID, unable to upload the attachments.")

        # If we had a single file, we could upload it with a single request that way:
        # `files={"FileUpload": (a.name, a.data)}`
        logger.debug("Submitting mail")
        r = post(url, data=fields, cookies=cookies)

        return r.text

    def _clean_attachments(self, upload_form, url, form_id, cookies):
        """ Remove all the attachments, received in the first ticket article so that they are not forwareded. """
        REMOVE_ALL_ATTACHMENTS = True
        def bs(): return BeautifulSoup(upload_form.text, features="html.parser")

        if REMOVE_ALL_ATTACHMENTS:
            for td in reversed(bs().find_all('a', {'class': 'AttachmentDelete'})): # why reversed? When deleted, others data-file-id shift down.
                self._remove_attachment(td['data-file-id'], form_id, url, cookies)
        elif self.parser.source_file:  # remove at least the attachment we have been split from (avoid duplicity)
            td = bs().find('td', {'class': 'Filename'}, lambda tag: tag.string == self.parser.source_file.name)
            if td:  # such attachment exists
                tr = td.parent
                data_file_id = tr.find('a', {'class': 'AttachmentDelete'})['data-file-id']
                self._remove_attachment(data_file_id, form_id, url, cookies)

    def _remove_attachment(self, data_file_id, form_id, url, cookies):
        logger.debug(f"Removing FileID={data_file_id} from the ticket article attachments") # XXX
        vv = post(url,
             params={"Action": "AjaxAttachment",
                     "Subaction": "Delete",
                     "FormID": form_id,
                     "ChallengeToken": self.parser.otrs_token,
                     "FileID": data_file_id
                     },
             cookies=cookies)

    @staticmethod
    def _check_record(record, lineno):
        valid = ('CONTACTS' in record)
        if not valid:
            print(" Line {}: Record missing CONTACTS field ".format(lineno))
        return valid

    @staticmethod
    def _check_response(response):
        if Config.is_testing():
            logger.info(" Response length:\n " + str(len(response)))

        mo = re_title.search(response)
        if not mo:
            logger.warning(" Unrecognized response")
            logger.error(response)
            return False

        # if Config.is_debug():
        #     logger.info("Got response\n" + response)

        title = mo.group(1)
        if 'Předat - Tiket - ' in title or 'Forward - Ticket - ' in title:
            # XX we are not sure sending succeeded. Ex: if we do not include "To" recipient, we land on the same page
            #   but no message will be send because the page will just state recipient is missing.
            #   However, this is not a priority, this mostly works. You can always use SMTP via --references flag.
            return True

        elif title == 'Login - OTRS':
            logger.warning("\n\n *** Not logged in or wrong session cookie ***")
        elif title in ('Fatal Error - Frontend -  OTRS', 'Fatal Error - Rozhraní -  OTRS'):
            logger.warning("\n\n *** Bad CSRF token ***")
        else:
            logger.warning(" Unrecognized response: " + title)
            logger.error(response)
        return False

    @staticmethod
    def ask_value(value, description=""):
        sys.stdout.write(f'Change {description} ({value})? [s]kip or paste it: ')
        t = input()
        if not t or t.lower() == "s":
            if not value:
                print("Attention, there should not be an empty value.")
            return value
        else:
            return t

    def assure_tokens(self):
        """ Check and update by dialog OTRS credentials """
        force = False
        while True:
            if (force
                    or not self.parser.otrs_id or not self.parser.otrs_num
                    or not self.parser.otrs_cookie or not self.parser.otrs_token or not self.parser.attachment_name):
                self.parser.otrs_id = self.ask_value(self.parser.otrs_id, "ticket url-id")
                self.parser.otrs_num = self.ask_value(self.parser.otrs_num, "ticket long-num")
                self.parser.otrs_cookie = self.ask_value(self.parser.otrs_cookie, "cookie")
                self.parser.otrs_token = self.ask_value(self.parser.otrs_token, "token")
                self.parser.attachment_name = self.ask_value(self.parser.attachment_name, "attachment name")

            if is_yes(f"Ticket id = {self.parser.otrs_id}, ticket num = {self.parser.otrs_num},"
                      f" cookie = {self.parser.otrs_cookie}, token = {self.parser.otrs_token},"
                      f" attachment_name = {self.parser.attachment_name}"
                      f"\nWas that correct?"):
                return True
            else:
                force = True
                continue

    def process(self, e: Envelope):
        def assure_str(c):
            return c if isinstance(c, str) else ";".join(c)

        # body must be HTML because OTRS treats it as such
        # (we might do a convey option to determine whether OTRS treats a text as plain or html)
        message = e.message()
        if not any(x for x in ("<br", "<b>", "<i>", "<p", "<img") if x in message):
            message = e.message().replace("\n", "<br>\n")

        fields = {
            "Action": "AgentTicketForward",
            "Subaction": "SendEmail",
            "TicketID": str(self.parser.otrs_id),
            "Email": e.from_().address,
            "From": assure_str(e.from_()),
            "To": assure_str(e.to()),  # mails can be delimited by comma or semicolon
            "Subject": e.subject(),
            "Body": message,
            "ArticleTypeID": "1",  # mail-external
            "ComposeStateID": "4",  # open
            "ChallengeToken": self.parser.otrs_token,
        }

        try:  # XX as of Python3.8, convert "try" to if m := Config.get("signkeyid", "OTRS"):
            if OTRS_VERSION == 6:
                fields["EmailSecurityOptions"] = Config.get("signkeyid", "OTRS")
            else:
                fields["SignKeyID"] = Config.get("signkeyid", "OTRS")
        except KeyError:
            pass

        if e.cc():
            fields["Cc"] = assure_str(e.cc())
        if e.bcc():
            fields["Bcc"] = assure_str(e.bcc())

        attachments = e.attachments()
        cookies = {('OTRSAgentInterface' if OTRS_VERSION == 6 else 'Session'): self.parser.otrs_cookie}

        if Config.is_testing():
            print(" **** Testing info:")
            print(' ** Fields: ' + str(fields))
            print(' ** Files: ', [(a.name, len(a.data)) for a in attachments])
            print(' ** Cookies: ' + str(cookies))

        host = Config.get("otrs_host", "OTRS")
        url = (host if "://" in host else f"https://{host}") + Config.get("baseuri", "OTRS")
        try:
            res = self._post_multipart(url,
                                       fields=fields,
                                       cookies=cookies,
                                       attachments=e.attachments())
        except Exception:
            import traceback
            logger.error(traceback.format_exc())
            logger.error(f"\nE-mail could not be send to the host {url} with the fields {fields}."
                         f" Are you allowed to send from this e-mail etc?")
            raise
        if not res or not self._check_response(res):
            print("Sending failure, see convey.log.")
            return False
        else:
            return True


class MailSenderSmtp(MailSender):
    smtp: smtplib.SMTP

    def start(self):
        try:
            self.smtp = smtplib.SMTP(Config.get("smtp_host", "SMTP"))
        except (gaierror, ConnectionRefusedError) as e:
            print("Can't connect to SMTP server", e)
            return False

    def stop(self):
        self.smtp.quit()

    def process(self, e: Envelope):
        e.smtp(self.smtp)
        return bool(e.send(True))



# Source file parsing
import csv
import datetime
import logging
import re
import subprocess
import time
from collections import defaultdict
from itertools import zip_longest
from json import dumps
from math import ceil
from operator import eq, ne
from pathlib import Path
from shutil import move
from typing import List, Optional, Tuple, Union
from sys import exit

from tabulate import tabulate

from .attachment import Attachment
from .contacts import Contacts
from .config import Config, get_terminal_size
from .definition import Settings
from .dialogue import Cancelled, is_yes, ask
from .field import Cell, Field
from .identifier import Identifier
from .informer import Informer
from .mail_draft import MailDraft
from .processor import Processor
from .types import Types, Type, Web, TypeGroup, Checker
from .whois import Whois

logger = logging.getLogger(__name__)


class Parser:
    is_split: bool
    "has already been split into many files (or we are just about to start file processing)"
    is_analyzed: bool
    "has already been analyzed"
    is_processable: bool
    "there are known actions we are ready to perform"
    is_processed: bool
    "has already been processed"
    attachments: List[Attachment]

    def __init__(self, source_file: Path = False, stdin=None, types=None, prepare=True):
        self.is_formatted = False
        self.is_repeating = False
        self.dialect = None
        "CSV dialect"
        self.has_header = None
        "CSV has header"
        self.is_pandoc = False
        "pandoc table format (second line to be skipped, a lot of spaces)"
        self.header: str = ""
        "if CSV has header, it's here so that Processor can take it"
        self.sample: List[str] = []
        "lines of the original file, including first line - possible header"
        self.sample_parsed: List[List[Cell]] = []
        """Values of the prepared output (ex re-sorted), always excluding header.
        Note that Field.compute_preview might append here a list instead of a str, hence the type Cell. """
        self.fields: List[Field] = []
        "CSV columns that will be generated to an output. Only parser.add_field is used to append to this list."
        self.first_line_fields: List[str] = []
        "CSV columns (equal to header if used) in the original file"
        self.types = []
        "field types of the columns as given by the user"
        self.settings: Settings = defaultdict(list)
        "processing settings"
        self.redo_invalids = Config.get("redo_invalids")
        # OTRS attributes to be linked to CSV
        self.otrs_cookie = False
        self.otrs_id = Config.get("ticketid", "OTRS")
        self.otrs_token = False
        self.otrs_num = Config.get("ticketnum", "OTRS")
        self.attachment_name = (source_file.name if source_file else "attachment.csv")
        if not Path(self.attachment_name).suffix:
            self.attachment_name += ".csv"
        self.ip_count_guess = None
        self.ip_count = None
        self.attachments = []
        "files created if splitting"
        self.queued_lines_count: int
        self.invalid_lines_count: int
        self.unknown_lines_count: int
        self.line_count = 0
        self.time_start = self.time_end = None
        # when called from another program we communicate through this stream rather than through a file
        # XX this is not used right now, convey is not at the moment connectable to other programs
        # see __init__.py at revision d2cf88f48409ca8cc5e229954df34836de884445
        self.external_stdout = None
        self.is_single_query = False
        "CSV processing vs single_query check usage"
        self.ranges = {}  # XX should be refactored as part of Whois
        self.ip_seen = {}  # XX should be refactored as part of Whois
        self.aggregation = defaultdict(dict)
        "set by processor [location file][grouped row][order in aggregation settings] = [sum generator, count]"
        self.refresh()
        self._reset(reset_header=False)
        self.selected: List[int] = []
        "list of selected fields col_i that may be in/excluded and moved in the menu"
        self.files_created = set()
        "files created with this parser, will not be rewritten but appended to if reprocessing lines"

        # load CSV
        self.source_file: Path = source_file or self.invent_file_str()
        self.stdin = []
        self.stdout = Config.get("stdout")
        """When accepting input from stdin and not saving the output into a file
          or when setting this to True,
          the output will be here.
        Config.get("output") - the file we should save to
        Config.get("stdout") - stdout used"""
        self.stdout_sample = None
        self.target_file = None
        self.saved_to_disk = None
        "has been saved to self.target_file"
        self.processor = Processor(self)
        self.informer = Informer(self)
        self.identifier = Identifier(self)

        if stdin:  # we're analysing an input text
            self.set_stdin(stdin)
        elif source_file:  # we're analysing a file on disk
            self.first_line, self.sample, self.is_pandoc = self.identifier.get_sample(self.source_file)
            self.lines_total, self.size = self.informer.source_file_len(self.source_file)
        self.set_types(types)
        # otherwise we are running a webservice which has no stdin nor source_file

        if prepare:
            self.prepare()

    def refresh(self):
        """
        Refresh dependency files – contact list Xwe need to tell the attachments if they are deliverable.
        """
        Contacts.init()
        Attachment.init(self)
        MailDraft.init(self)

    def prepare(self):
        if self.size == 0:
            print("Empty contents.")
            exit()
        self.prepare_target_file()

        # check if we are parsing a single cell
        if self.stdin:
            seems = True

            def join_base(s):
                return "".join(s).replace("\n", "").replace("\r", "")

            def join_quo(s):
                return "\n".join(s).replace("=\n", "").replace("=\r", "").replace("\n", r"\n").replace("\r", r"\r")

            len_ = len(self.sample)
            if len_ > 1 and re.search("[^A-Za-z0-9+/=]", join_base(self.sample)) is None:
                # in the sample, there is just base64-chars
                s = join_base(self.stdin)
                seems = False
                try:
                    if Checker.is_base64(s):  # all the input is base64 decodable
                        seems = True
                        self.set_stdin([s])
                except ValueError:
                    pass
            elif len_ > 1 and re.search("=[A-Z0-9]{2}", join_quo(self.sample)):
                s = join_quo(self.stdin)

                seems = False
                if Checker.is_quopri(s):
                    seems = True
                    self.set_stdin([s])
            elif not len_ or len_ > 1:
                seems = False

            if Config.get("single_query") or (seems and Config.get("single_query") is not False):
                # identify_fields some basic parameters
                self.add_field([Field(self.stdin[0])])  # stdin has single field
                self.dialect = False
                self.has_header = False
                if self.identifier.identify_fields(quiet=True):
                    # tell the user what type we think their input is
                    # access the detection message for the first (and supposedly only) field
                    detection = self.get_fields_autodetection(False)[0][1]
                    if not detection and not Config.get("adding-new-fields"):
                        # this is not a single cell despite it was probable, let's continue input parsing
                        logger.info("We could not parse the input text easily.")
                    else:
                        if not detection:
                            # we are adding new fields - there is a reason to continue single processing
                            logger.info("Input value seems to be plaintext.")
                        elif Config.is_quiet() and Config.get("single_detect", get=bool):
                            print(detection)  # print out detection even if we are quiet because explicitly asked for
                        else:
                            logger.info(f"Input value {detection}\n")
                        self.is_single_query = True
                        return self
                if Config.get("single_query"):
                    logger.info("Forced single processing")
                    self.fields[0].possible_types = {Types.plaintext: 1}
                    self.is_single_query = True
                    return self

        # we are parsing a CSV file
        self.sample_parsed = [x for x in csv.reader(self.sample)]
        self.informer.sout_info()
        try:
            # Dialog to obtain basic information about CSV - delimiter, header
            self.dialect, self.has_header, seems_single = self.identifier.guess_dialect(self.sample)
            uncertain = False

            l = []
            if self.is_pandoc:
                l.append("Pandoc table format detected (header were underlined by -------)")
            if Config.get("delimiter", "CSV"):
                self.dialect.delimiter = Config.get("delimiter", "CSV")
                l.append(f"Delimiter character set: '{self.dialect.delimiter}'")
                self.dialect.delimiter = self.dialect.delimiter.replace(r"\t", "\t").replace("TAB", "\t").replace("tab",
                                                                                                                  "\t")
            else:
                uncertain = True
                s = "proposed" if seems_single else "found"
                l.append(f"Delimiter character {s}: '{self.dialect.delimiter}'")

            if Config.get("quote_char", "CSV"):
                self.dialect.quotechar = Config.get("quote_char", "CSV")
                l.append(f"Quoting character set: '{self.dialect.quotechar}'")
            else:
                uncertain = True
                l.append(f"Quoting character: '{self.dialect.quotechar}'")

            if Config.get("header", "CSV") is not None:
                self.has_header = Config.get("header", "CSV")
            else:
                uncertain = True
                l.append(f"Header is present: " + ("yes" if self.has_header else "not used"))

            if Config.get("yes"):
                uncertain = False
            else:
                print("\n".join(l))

            if uncertain and not is_yes("\nCould you confirm this?"):
                while True:
                    s = "What is delimiter " + (
                        f"(default '{self.dialect.delimiter}')" if self.dialect.delimiter else "") + ": "
                    self.dialect.delimiter = input(s) or self.dialect.delimiter
                    if len(self.dialect.delimiter) != 1:
                        print("Delimiter must be a 1-character string. Invent one (like ',').")
                        continue
                    s = "What is quoting char " + (
                        f"(default '{self.dialect.quotechar}')" if self.dialect.quotechar else "") + ": "
                    self.dialect.quotechar = input(s) or self.dialect.quotechar
                    break
                self.dialect.quoting = csv.QUOTE_NONE if not self.dialect.quotechar else csv.QUOTE_MINIMAL
                if not is_yes("Header " + ("" if self.has_header else "not found; ok?")):
                    self.has_header = not self.has_header
            self.first_line_fields = csv.reader([self.first_line], skipinitialspace=self.is_pandoc,
                                                dialect=self.dialect).__next__()
            self.reset_settings()
        except Cancelled:
            print("Cancelled.")
            exit()
        self.informer.sout_info()
        self.is_formatted = True  # delimiter and header has been detected etc.
        return self

    def get_fields_autodetection(self, append_values=True):
        """ returns list of tuples [ (field, detection str), ("Url", "url, hostname") ]
        :type append_values: bool Append sample values to the informative result string.
        """
        fields = []
        for col, field in enumerate(self.fields):
            s = ""
            if field.is_new:
                s = f"computed from: {field.source_type}"
            elif field.type:
                s = field.type
            elif field.possible_types:
                types = field.possible_types
                if Types.any_ip in types and (Types.ip in types or Types.port_ip in types):
                    del types[Types.any_ip]
                if Types.url in types and Types.wrong_url in types:
                    del types[Types.url]
                s = f"detected: {', '.join((str(e) for e in types))}"
                if append_values:
                    s += " – values: " + ", ".join(field.get_samples(3))
            fields.append((field, s))
        return fields

    def add_field(self, replace: Optional[List[Field]] = None, append: Optional[Field] = None):
        """
        :param replace: Replace fields by the new field list.
        :param append: Append new field to the current list.
        """
        fields = []
        if replace:
            self.fields = []
            fields = replace
        if append:
            fields += [append]

        for f in fields:
            f.col_i_original = f.col_i = len(self.fields)
            if self.types and len(self.types) > f.col_i:
                f.type = self.types[f.col_i]
            f.set_parser(self)  # register Field to this parser
            self.fields.append(f)
        return self

    def set_types(self, types):
        if types:
            try:
                self.types = [getattr(Types, t) for t in types.split(",")]
            except AttributeError:
                print(f"Unknown type amongst: {types}")
                exit()
        else:
            self.types = []
        return self

    def get_computed_fields(self):
        for f in self.fields:
            if f.is_new:
                yield f

    def set_stdin(self, stdin):
        self.stdin = stdin
        self.lines_total = self.size = len(self.stdin)
        if self.size:
            self.first_line, self.sample = self.stdin[0], self.stdin[:7]
        return self

    def run_single_query(self, json=False):
        """ Print out meaningful details about the single-value contents.
        :param json: If true, returns json.
        """
        # prepare the result variables
        rows = []
        data = {}
        Web.init()

        def append(target_type, val):
            """
            :type target_type: Field|Type If Field, it is added on purpose. If Type, convey just tries to add all.
            """
            data[str(target_type.name)] = [str(v) for v in val] if type(val) is list else (str(val) if val else "")
            if type(val) is list and len(val) == 1:
                val = val[0]
            elif val is None or val == [] or (type(val) is str and val.strip() == ""):
                if Config.is_verbose() or not hasattr(target_type, "group" or target_type.group == TypeGroup.general):
                    # if not verbose and target_type is Type and no result, show just TypeGroup.general basic types
                    val = "×"
                else:
                    return
            else:
                val = str(val)
            rows.append([str(target_type.name), val])

        # get fields and their methods to be computed
        fields = [(f, f.get_methods()) for f in self.fields if f.is_new]  # get new fields only
        if fields and not any(1 for f in fields if f[0].is_chosen):
            print("All fields are excluded. Do not know what to display.")
            return

        custom_fields = True
        if not fields:  # transform the field by all known means
            custom_fields = False
            if self.fields[0].type.is_plaintext_derivable:
                # Ex: when input is quoted_printable, we want plaintext only,
                # not plaintext re-encoded in base64 by default.
                # (It is easy to specify that we want base64 by `--field base64`)
                types = [Types.plaintext]
            else:  # loop all existing methods
                types = Types.get_computable_types(ignore_custom=True)
            for target_type in types:
                if target_type in Config.get("single_query_ignored_fields", "FIELDS", get=list):
                    # do not automatically compute ignored fields
                    continue
                elif target_type.group == TypeGroup.custom:
                    continue
                fitting_type = self.identifier.get_fitting_type(0, target_type)
                if fitting_type:
                    methods = self.identifier.get_methods_from(target_type, fitting_type, None)
                    if methods:
                        fields.append((target_type, methods))

        method_cache = {}
        single_query_value = "\n".join(self.stdin)
        row = [single_query_value]
        for field, methods in fields:
            if type(field) is Type:
                val = single_query_value
            else:
                val = row[field.source_field.col_i]
            try:
                for i, l in enumerate(methods):
                    if (repr((val, methods[:i + 1]))) in method_cache:
                        val = method_cache[(repr((val, methods[:i + 1])))]
                        continue
                    if isinstance(val, list):
                        # resolve all items, while flattening any list encountered
                        val = [y for x in (l(v) for v in val) for y in (x if type(x) is list else [x])]
                    else:
                        val = l(val)
                    # We cache this value so that it will not be recomputed when crawling the same path on the graph again.
                    # Ex: `hostame → ip → country` and `hostname → ip → asn` will not call method `hostname → ip` twice (for each).
                    method_cache[(repr((val, methods[:i + 1])))] = val
            except Exception as e:
                val = str(e)
            row.append(val)
            if type(field) is Type or field.is_chosen:
                append(field, val)

        # output in text, json or file
        if Config.get("output"):
            logger.info(f"Writing to {self.target_file}...")
            self.target_file.write_text(dumps(data))
        if json:
            return dumps(data)
        elif Config.is_quiet() and len(rows) == 1:
            print(rows[0][1])
        else:
            # pad to the screen width
            width = get_terminal_size()[1]
            if rows and width:
                # size of terminal - size of the longest field name + 10 column space
                width -= max(len(row[0]) for row in rows) + 10
                for i, row in enumerate(rows):
                    val = row[1]
                    if width and len(str(val)) > width:  # split the long text by new lines
                        row[1] = "\n".join([val[i:i + width] for i in range(0, len(val), width)])
            if not rows and custom_fields:
                s = ", ".join([str(f) for f in fields])
                print(f"Cannot compute {s}")
            else:
                print(tabulate(rows, headers=("field", "value")))

    def reset_whois(self, hard=True, assure_init=False, slow_mode=False, unknown_mode=False):
        """

        :type assure_init: Just assure the connection between picklable Parser and current Whois class.
        """
        if not assure_init:
            if hard:
                self.whois_stats = defaultdict(int)
                self.ranges = {}
                self.ip_seen = {}
        Whois.init(self.whois_stats, self.ranges, self.ip_seen, self.stats, slow_mode=slow_mode,
                   unknown_mode=unknown_mode)

    def reset_settings(self):
        self.settings = defaultdict(list)
        # when resetting settings, we free up any finished aggregation
        # (because informer wants to display it but the self.parser.settings["aggregate"] is gone
        self.aggregation = defaultdict(
            dict)  # self.aggregation[location file][grouped row][order in aggregation settings] = [sum generator, count]
        self.sample_parsed = [x for x in
                              csv.reader(self.sample[slice(1 if self.has_header else 0, None)],
                                         skipinitialspace=self.is_pandoc,
                                         dialect=self.dialect)]
        self.add_field([Field(f) for f in self.first_line_fields])
        self.identifier.identify_fields()

    def _reset_output(self):
        self.line_count = 0
        self.velocity = 0

    def _reset(self, hard=True, reset_header=True):
        """ Reset variables before new analysis.
        @type reset_header: False if we are in the constructor and added fields is not ready yet.
        """
        self.stats = defaultdict(set)
        Attachment.reset(self.stats)
        self.queued_lines_count = self.invalid_lines_count = self.unknown_lines_count = 0
        # self.aggregation[location file][grouped row][order in aggregation settings] = [sum generator, count]
        self.aggregation = defaultdict(dict)

        if reset_header:
            class Wr:  # very ugly way to correctly get the output from csv.writer
                def write(self, row):
                    self.written = row

            wr = Wr()
            cw = csv.writer(wr, dialect=self.settings["dialect"])
            cw.writerow([f for f in self.fields if f.is_chosen])
            self.header = wr.written
        self._reset_output()
        # self.get_sample_values()  # assure sout_info would consume a result from duplicate_row

        self.time_start = None
        self.time_end = None
        self.is_analyzed = False
        self.is_split = False
        self.is_processable = False
        self.is_processed = False
        self.attachments.clear()
        self.reset_whois(hard=hard)

    def prepare_target_file(self):
        if not self.settings["split"] and self.settings["split"] != 0:  # 0 is a valid column
            self.is_split = False
            self.target_file = self.invent_file_str()
        else:
            self.is_split = True
            self.target_file = None
        return self.target_file

    def invent_file_str(self):
        l = []
        se = self.settings
        if se["filter"]:
            l.append("filter")
        if se["unique"]:
            l.append("uniqued")
        if se["dialect"] and (se["dialect"].delimiter != self.dialect.delimiter
                              or se["dialect"].quotechar != self.dialect.quotechar):
            l.append("dialect")
        if se["header"] is False:
            l.append("noheader")
        if [f for f in self.fields if not f.is_chosen]:
            l.append("shuffled")
        for f in se["add"]:
            # XX get external function name
            # if f.type == Types.external:
            #     l.append(str(f))
            # else:
            l.append(str(f))
        agg = se["aggregate"]
        if agg:
            if agg.group_by is not None:
                l.append(f"{agg.group_by}-grouped")
            for fn, col in agg.actions:
                l.append(f"{fn.__name__}-{col}")
        if hasattr(self, "source_file"):
            l.insert(0, self.source_file.stem)
            target_file = f"{'_'.join(l)}{self.source_file.suffix}"
        else:
            target_file = f"output_{time.strftime('%Y-%m-%d %H:%M:%S')}.csv"
        output = Config.get("output")
        return Path(str(output)) if output else Path(Config.get_cache_dir(), target_file)

    def run_analysis(self, autoopen_editor=None):
        """ Run main analysis of the file.
        :type autoopen_editor: bool May shadow config file value "autoopen_editor"
        """
        self.refresh()
        self._reset(hard=False)

        if (autoopen_editor or autoopen_editor is None) and Config.get("autoopen_editor") and self.is_split:
            Contacts.mail_draft["local"].edit_text(blocking=False)
            Contacts.mail_draft["abroad"].edit_text(blocking=False)

        self.time_start = datetime.datetime.now().replace(microsecond=0)
        self.prepare_target_file()
        self.processor.process_file(self.source_file, rewrite=True, stdin=self.stdin)
        self.time_end = datetime.datetime.now().replace(microsecond=0)
        self.lines_total = self.line_count  # if we guessed the total of lines, fix the guess now
        self.is_analyzed = True
        self.is_processed = True
        self.informer.sout_info()
        if self.invalid_lines_count:
            self.resolve_invalid()

        if self.unknown_lines_count:
            self.resolve_unknown()

        if self.queued_lines_count and Config.get("lacnic_quota_resolve_immediately", "FIELDS") is not False:
            self.resolve_queued(Config.get("lacnic_quota_resolve_immediately", "FIELDS"))

        self.line_count = 0

    def _guess_ip_count(self):
        """ Determine how many IPs there are in the file.
        XX not used and not right (doesnt implement dialect but only delimiter) (doesnt implement stdin instead of source_file)
        """
        if self.urlColumn is None:
            try:
                max = 100000
                i = 0
                ipSet = set()
                fraction = None
                with open(self.source_file, 'r') as csvfile:
                    for line in csvfile:
                        i += 1
                        if self.has_header and i == 1:
                            continue
                        ip = line.split(self.dialect.delimiter)[self.ipColumn].strip()
                        ipSet.add(ip)
                        if i == (max - 1000):
                            fraction = len(ipSet)
                        if i == max:
                            break
                if i != max:
                    self.ip_count = len(ipSet)
                    print("There are {} IPs.".format(self.ip_count))
                else:
                    delta = len(ipSet) - fraction  # determine new IPs in the last portion of the sample
                    self.ip_count_guess = len(ipSet) + ceil((self.lines_total - i) * delta / i)
                    print(
                        "In the first {} lines, there are {} unique IPs. There might be around {} IPs in the file.".format(
                            i, len(ipSet), self.ip_count_guess))
            except Exception:
                print("Can't guess IP count.")

    def _resolve_again(self, path, basename, var, key, reprocess_method, slow_mode=False, unknown_mode=False):
        count = getattr(self, var)  # ex: `count = self.invalid_lines_count`
        setattr(self, var, 0)
        self.reset_whois(assure_init=True, slow_mode=slow_mode, unknown_mode=unknown_mode)
        temp = str(path) + ".running.tmp"
        try:
            move(path, temp)
        except FileNotFoundError:
            input("File {} not found, maybe resolving was run in the past and failed. Please rerun again.".format(path))
            return False

        self._reset_output()
        lines_total, size = self.lines_total, self.size
        self.lines_total, self.size = self.informer.source_file_len(temp)
        if basename in self.files_created:
            self.files_created.remove(basename)  # this file exists no more, if recreated, include header
        dialect_tmp = self.dialect
        self.dialect = self.settings["dialect"]
        # XX missing start time reset here
        self.processor.process_file(temp)
        self.dialect = dialect_tmp
        Path(temp).unlink()
        self.lines_total, self.size = lines_total, size
        self._reset_output()
        self.informer.sout_info()

        if getattr(self, var):
            solved = count - getattr(self, var)
            print(f"\nNo {key} row resolved." if solved == 0 else f"\nOnly {solved}/{count} {key} rows were resolved.")
            reprocess_method()

    def stat(self, key):
        """ self.stats reader shorthand (no need to call len(...))"""
        return len(self.stats[key])

    def resolve_unknown(self):
        """ Process all prefixes with unknown abusemails.

        When split processing, unknown file is always generated; when single file processing,
        it is generated only if whois_reprocessable_unknown is True, otherwise cells remain empty

        It is not trivial to determine which IPs are really marked as unknowns.
        If we are splitting by abusemail, this may do. But if we are splitting by incident-contact,
        some of the IP with unknown abusemails are deliverable through csirtmail.
        Whois module does statistics but is not able to determine whether we use abusemail or incident-contact.

        That is the reason we are saying 'up to': "There are ... lines with up to ... IPs".
        """

        # (self.is_split or Config.get("whois_reprocessable_unknown", "FIELDS", get=bool)) \
        #    and (self.stats["prefix_local_unknown"] or self.stats["prefix_abroad_unknown"]):

        if not self.unknown_lines_count:
            # self.stats["ip_local_unknown"] and not self.stats["ip_abroad_unknown"]:
            input("No unknown abusemails. Press Enter to continue...")
            return

        l = []
        if self.stats["ip_local_unknown"]:
            l.append(f"{self.stat('ip_local_unknown')} IPs in {self.stat('prefix_local_unknown')} unknown prefixes")
        if self.stats["ip_abroad_unknown"]:
            l.append(
                f"{self.stat('ip_abroad_unknown')} IPs in {self.stat('prefix_abroad_unknown')} unknown abroad prefixes")
        s = f"There are {self.unknown_lines_count} lines with up to {' and '.join(l)}. Should I proceed additional search" \
            f" for these {self.stat('ip_local_unknown') + self.stat('ip_abroad_unknown')} IPs?"
        if not is_yes(s):
            return

        path = Path(Config.get_cache_dir(), Config.UNKNOWN_NAME)
        [self.stats[f"{a}_{b}_unknown"].clear() for a in ("ip", "prefix") for b in
         ("local", "abroad")]  # reset the stats matrix
        # res = self._resolve_again(path, Config.UNKNOWN_NAME, unknown_mode=True)
        self._resolve_again(path, Config.UNKNOWN_NAME, "unknown_lines_count", "unknown", self.resolve_unknown,
                            unknown_mode=True)
        Whois.unknown_mode = False
        # if res is False:
        #     return False

    def resolve_queued(self, force=False):
        count = self.queued_lines_count
        if not count:
            return True
        print(f"There are {count} queued rows")
        if Whois.quota.remains():
            print(f"We will have to wait {Whois.quota.remains()} s before LACNIC quota is over.")
        if not (force or is_yes(f"Reanalyse them?")):
            return False
        if Whois.quota.remains():
            print(f"Waiting till {Whois.quota.time()} or Ctrl-C to return", end="")
            try:
                while True:
                    time.sleep(1)
                    print(".", end="", flush=True)
                    if not Whois.quota.remains():
                        break
            except KeyboardInterrupt:
                return
            print(" over!")

        path = Path(Config.get_cache_dir(), Config.QUEUED_NAME)
        Whois.queued_ips.clear()

        self._resolve_again(path, Config.QUEUED_NAME, "queued_lines_count", "queued", self.resolve_queued,
                            slow_mode=True)
        # self.queued_lines_count = 0
        # res = self._resolve_again(path, Config.QUEUED_NAME, slow_mode=True)
        # if res is False:
        #     return False
        # if self.queued_lines_count:
        #     solved = count - self.queued_lines_count
        #     if solved == 0:
        #         s = "No queued row resolved."
        #     else:
        #         s = f"Only {solved}/{count} queued rows were resolved."
        #     print("\n" + s)
        #     self.resolve_queued()

    def resolve_invalid(self):
        """ Process all invalid rows. """
        if Config.get("yes", get=bool) and Config.is_quiet():
            return False
        if not self.invalid_lines_count:
            input("No invalid rows. Press Enter to continue...")
            return

        path = Path(Config.get_cache_dir(), Config.INVALID_NAME)
        print("There are {0} invalid rows".format(self.invalid_lines_count))
        while True:
            try:
                with open(path, 'r') as f:
                    for i, row in enumerate(f):
                        print(row.strip())
                        if i > 5:
                            break
            except FileNotFoundError:
                input("File {} not found, maybe resolving was run in the past and failed. Please rerun again.".format(
                    path))
                return False
            if Config.get("yes", get=bool):
                res = "n"
            else:
                s = "Open the file in text editor (o) and make the rows valid, when done, hit y for reanalysing them," \
                    " or hit n for ignoring them. [o]/y/n "
                res = ask(s)
            if res == "n":
                return False
            elif res == "y":
                break
            else:
                print("Opening the editor...")
                subprocess.Popen(['xdg-open', path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        self._resolve_again(path, Config.INVALID_NAME, "invalid_lines_count", "invalid", self.resolve_invalid)
        # invalids = self.invalid_lines_count
        # self.invalid_lines_count = 0
        # if self._resolve_again(path, Config.INVALID_NAME) is False:
        #     return False
        # if self.invalid_lines_count:
        #     solved = invalids - self.invalid_lines_count
        #     if solved == 0:
        #         s = "No invalid row resolved."
        #     else:
        #         s = f"Only {solved}/{invalids} invalid rows were resolved."
        #     print("\n" + s)
        #     self.resolve_invalid()

    def resort(self, chosens=[]):
        """
        :type chosens: List[str] Column IDs in str, ex: "1" points to column index 0
        """
        for f in self.fields:
            f.is_chosen = False
        l = [str(c + 1) for c in range(len(self.fields))]
        excluded = [f for f in l if f in set(l) - set(chosens)]
        transposed = list(zip(*self.sample_parsed))
        sp = []
        for i, (col, is_chosen) in enumerate([(col, True) for col in chosens] + [(col, False) for col in excluded]):
            col_i = self.identifier.get_column_i(col, check="to be re-sorted")
            self.fields[col_i].col_i = i
            if self.fields[col_i].is_new:
                sp.append([None] * len(self.sample_parsed))
            else:
                sp.append(transposed[col_i])
            self.fields[col_i].is_chosen = is_chosen
        self.fields = sorted(self.fields, key=lambda f: f.col_i)
        self.sample_parsed = list(map(list, zip(*sp)))  # force list preventing tuples

    def move_selection(self, direction=0, move_contents=False):
        """ Move cursor or whole columns
            :type move_contents: bool Move whole columns, not only cursor.
            :type direction: int +1 right, -1 left
        """
        selected = [f for f in self.fields if f.is_selected]
        if not selected:  # initial value is on an either border
            i = direction - 1 if direction > 0 else direction
            self.fields[i % len(self.fields)].is_selected = True
            if not move_contents:
                return
        if move_contents:
            for f in [f for f in self.fields if f.is_selected]:
                f.move(direction)
            self.is_processable = True
        else:
            for f in selected:  # deselect previous ones
                f.is_selected = False
            for f in selected:  # select new ones
                self.fields[(f.col_i + direction) % len(self.fields)].is_selected = True

    def get_sample_values(self):
        settings = self.settings
        unique_sets = defaultdict(set)

        rows = []  # nice table formatting
        full_rows = []  # formatting that optically matches the Sample above
        for line in self.sample_parsed:
            row: List[Tuple[str,Field]] = []
            for field, cell in zip_longest(self.fields, line):
                if field is None:
                    # when lines have different length, field may be none
                    # XX should the processing be stopped?
                    field = Field("plaintext")
                if cell is None:
                    cell = field.compute_preview(line)
                # suppress new lines in preview; however while processing,
                # new lines are printed and this may render the CSV unloadable
                row.append((str(cell).replace("\n", r"\n"), field))

            # check if current line is filtered out
            line_chosen = True
            for include, col_i, val in settings["filter"]:
                # (include, col, value), ex: [(True, 23, "passed-value"), ...]
                if (ne if include else eq)(val, row[col_i][0]):
                    line_chosen = False

            if settings["unique"]:  # unique columns
                for col_i in settings["unique"]:  # list of uniqued columns [2, 3, 5, ...]
                    if row[col_i][0] in unique_sets[col_i]:  # skip line
                        line_chosen = False
                        break
                else:  # do not skip line
                    for col_i in settings["unique"]:
                        unique_sets[col_i].add(row[col_i][0])

            # colorize the line
            g = lambda short: (field.color(cell, short, line_chosen) for cell, field in row)
            rows.append([*g(True)])
            full_rows.append([*g(False)])
        return full_rows, rows

    def get_similar(self, fields2: Union[Field, List[Field]]):
        """ Recommend which fields might be similar in different parsers.
            Based on the same column type.

            (In the future, we might implement comparing the similarity by the parser sample values too.)
        """
        if isinstance(fields2, Field):
            fields2 = [fields2]
        return [f1 for f1 in self.fields if f1.type for f2 in fields2 if f1.type == f2.type]



    def __getstate__(self):
        state = self.__dict__.copy()
        del state['informer']
        del state['processor']
        del state['identifier']
        del state['ip_seen']  # delete whois dicts
        del state['ranges']
        # counters generators may be removed (their state is not jsonpicklable)
        # however that means that when re-resolving after main processing, generators counts will reset
        # See comment in the Aggregation class, concerning generator serialization.
        # counters[location file][grouped row][order in aggregation settings] = [sum generator, count]
        for l in state['aggregation'].values():
            for g in l.values():
                for o in g:
                    o[0] = None
        state['dialect'] = self.dialect.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.informer = Informer(self)
        self.processor = Processor(self, rewrite=False)

        # input CSV dialect
        self.dialect = csv.unix_dialect()
        for k, v in state["dialect"].items():
            setattr(self.dialect, k, v)

        # output CSV dialect
        d = self.settings["dialect"]
        if d:
            self.settings["dialect"] = csv.unix_dialect()
            for k, v in d.items():
                setattr(self.settings["dialect"], k, v)

        self.identifier = Identifier(self)
        self.ranges = {}
        self.ip_seen = {}



import csv
import re
from pathlib import Path
from typing import Dict

from .config import Config, get_path
from .mail_draft import MailDraft


class Contacts:
    mail2cc: Dict[str, str]
    country2mail: Dict[str, str]
    mail_draft: Dict[str, MailDraft]

    @classmethod
    def init(cls):
        """
        Refreshes list of abusemails (for Cc of the mails in the results) (config key contacts_cc)
        and csirtmails (country contact) (config key contacts_abroad)
        """
        cls.mail_draft = {"local": MailDraft(Config.get("mail_template")),
                          "abroad": MailDraft(Config.get("mail_template_abroad"))}
        cls.mail2cc = cls._update("contacts_cc")
        cls.country2mail = cls._update("contacts_abroad")

    @staticmethod
    def get_domains(mail: str):
        """ mail = mail@example.com;mail2@example2.com -> [example.com, example2.com] """
        try:
            # return set(re.findall("@([\w.]+)", mail))
            return set([x[0] for x in re.findall("@(([A-Za-z0-9-]{1,63}\.)+[A-Za-z]{2,6})", mail)])
        except AttributeError:
            return []

    @staticmethod
    def _update(key: Dict[str, str]) -> object:
        """ Update info from an external CSV file. """
        file = get_path(Config.get(key))
        if not Path(file).is_file():  # file with contacts
            print("(Contacts file {} not found on path {}.) ".format(key, file))
            input()
            return {}
        else:
            with open(file, 'r') as f:
                reader = csv.reader(f)
                next(reader)  # skip header row
                try:
                    rows = {rows[0]: rows[1] for rows in reader}
                except IndexError:
                    raise IndexError(f"Error while loading file {file}")
                return rows


from csv import Dialect
from sys import version_info
from typing import Callable, Dict, TYPE_CHECKING, List, Optional, Tuple, Union

from .action import AggregateAction, MergeAction
from .field import Field

try:
    from typing import TypedDict
except ImportError:  # remove as of Python 3.7
    if not TYPE_CHECKING:  # that way, IDE displays correct values and Python 3.7 still works
        TypedDict = Dict

Unique = Optional[List[int]]
Filter = Optional[List[Tuple[bool, int, str]]]

class Settings(TypedDict):
    # XX Convert more to Actions, get rid of complex fields

    add: Optional[List[Field]]
    """  XX handle this old doc
            "add": new_field:Field,
                source_col_i:int - number of field to compute from,
                fitting_type:Field - possible type of source ,
                custom:tuple - If target is a 'custom' field, we'll receive a tuple (module path, method name).
    """
    aggregate: Optional[AggregateAction]
    dialect: Dialect
    "always present (set in controller just after parser.prepare()), output CSV dialect"
    filter: Filter
    merge: Optional[List[MergeAction]]
    split: Optional[int]
    unique: Unique
    header: Optional[bool]
    """ True if input CSV has header and output CSV should have it too.
        False if either input CSV has not header or the output CSV should omit it."""

    # used and set only in Processor
    addByMethod: Optional[List[Tuple[str, int, List[Callable]]]]
    f_pre: Filter
    f_post: Filter
    target_file: Union[None, int, str]
    u_pre: Unique
    u_post: Unique
    merging: Optional[bool]



import logging
from collections import defaultdict

logger = logging.getLogger(__name__)


class Graph:
    def __init__(self):
        self.nodes = set()
        self.edges = defaultdict(list)
        self.distances = {}

    def clear(self):
        self.__init__()

    def add_node(self, value):
        self.nodes.add(value)

    def add_edge(self, to_node, from_node, distance=1):
        self.edges[from_node].append(to_node)
        self.edges[to_node].append(from_node)
        self.add_node(from_node)
        self.add_node(to_node)
        self.distances[(from_node, to_node)] = distance

    # @lru_cache(maxsize=512) XX is it a good performance tip?
    # XX what if we cache every encountered path instead?
    def dijkstra(self, target, start=None, ignore_private=False):
        """
        Performs Dijkstra's algorithm and returns
            [start node, ... , target node] or False (if start specified)
        OR
            {"node": distance to target, ...} (lower is better, sorted from lower)

            :param target: Type
            :type start: [] or False if no path exists
            :type ignore_private: bool True - do not return private notes within results (ex. whois). Not applicable with `start`.
        """
        visited = {target: 0}
        tree = {}

        nodes = set(self.nodes)

        while nodes:
            min_node = None
            for node in nodes:
                if node in visited:
                    if min_node is None:
                        min_node = node
                    elif visited[node] < visited[min_node]:
                        min_node = node

            if min_node is None:
                break

            nodes.remove(min_node)
            current_weight = visited[min_node]

            for edge in self.edges[min_node]:
                try:
                    weight = current_weight + self.distances[(min_node, edge)]
                    if edge not in visited or weight < visited[edge]:
                        visited[edge] = weight
                        tree[edge] = min_node
                except KeyError:
                    pass

        if start:
            start = start.computing_start
            path = [start]
            pivot = start
            i = 0
            while pivot != target:
                i += 1
                try:
                    pivot = tree[pivot]
                except KeyError:  # path does not exist
                    return False
                path.append(pivot)

                if i > 10:
                    logger.error("cycles", path, target)
                    break
            return path

        if ignore_private:
            for node in list(visited):
                if node.is_private:
                    del visited[node]
        distance_from_type = {}
        for (k, v) in visited.items():
            if v > 0:  # skip the same node
                for b in k.equals:
                    distance_from_type[b] = v
        return distance_from_type



from __future__ import annotations
import binascii
import logging
from base64 import b64decode
from pathlib import Path

from colorama import Fore
from envelope import Envelope
from jinja2 import Template, exceptions
from typing import TYPE_CHECKING

from .config import Config, get_path, edit

if TYPE_CHECKING:
    from .attachment import Attachment

PLAIN_DELIMITER = "\n"

HTML_DELIMITER = "<br>\n"

logger = logging.getLogger(__name__)


def is_html(s):
    return any(x in s for x in ("<br", "<p"))


class MailDraft:
    """ Mail management data structure """
    def __init__(self, filename):
        self.text = False
        self.template_file = get_path(filename)
        self.mail_file = Path(Config.get_cache_dir(), filename)
        self.file_assured = False  # we tried to load the e-mail text from the template and with the use of CLI flags

    @classmethod
    def init(cls, parser):
        cls.parser = parser

    def get_mail_preview(self) -> str:
        l = self.get_envelope().preview().splitlines()
        s = "\n".join(l[:40])
        if len(l) > 40:
            s += "\n..."
        return Fore.CYAN + s + Fore.RESET

    @staticmethod
    def _decode_text(t):
        base64_text = "data:text/plain;base64,"
        if t.startswith(base64_text):
            try:
                return b64decode(t[len(base64_text):]).decode("utf-8", "ignore")
            except binascii.Error:
                raise RuntimeError(f"Cannot decode text and create e-mail template: {t}")
        return t

    def edit_text(self, blocking=True):
        """ Opens file for mail text to GUI editing.
            Create the e-mail file from the template if had not existed before.
            If --subject, --body and --reference flags are used, they get re-merged into the template.
        """
        self._assure_file()
        edit(self.mail_file, mode=2, blocking=blocking)

    def _assure_file(self):
        """ Create the e-mail file from the template if had not existed before.
            If --subject, --body and --reference flags are used, they get re-merged into the template.
            :rtype: bool True if the e-mail file was newly created.
            """
        if self.file_assured:
            return False

        just_created = False

        # enrich e-mail text using the template
        if not self.mail_file.is_file():
            # t = Path(self.template_file).read_text()
            try:
                self.mail_file.write_text(self.template_file.read_text())
                logger.info(f"Writing from template {self.template_file}")
                just_created = True
            except FileNotFoundError:
                logger.warning(f"Template file {self.template_file} not found")

        # enrich e-mail text using the CLI variables `body`, `subject`, `references`
        subject = Config.get("subject")
        body = Config.get("body")
        references = Config.get("references")
        if subject or body or references:
            e = Envelope.load(self.mail_file).date(False)
            if subject:
                e.subject(self._decode_text(subject))
            if body:
                # reset loaded message from the template
                # to prevent the situation when template message is considered a text/plain alternative
                # and `--body` message is considered as a text/html alternative.
                # XX envelope might have method to delete message, like
                #  .message(False), .message(""), .message(text, alternative="replace")
                # We rather join template with the current message.
                template_message = e.message()
                message = self._decode_text(body).replace(r"\n", PLAIN_DELIMITER)
                if template_message:
                    # add / remove HTML from the template message
                    if is_html(template_message) and not is_html(message):
                        template_message = template_message.replace("<br>", PLAIN_DELIMITER)
                    elif not is_html(template_message) and is_html(message):
                        template_message = template_message.replace(PLAIN_DELIMITER, HTML_DELIMITER)
                    delimiter = HTML_DELIMITER if is_html(message) else PLAIN_DELIMITER
                    message = template_message + delimiter * 2 + message

                e.message("", alternative="auto") \
                    .message("", alternative="plain") \
                    .message("", alternative="html") \
                    .message(message)
            if references:
                if not (references.startswith("<") and references.endswith(">")):
                    references = f"<{references}>"
                e.header("Reference", references)
                e.bcc(Config.get("email_from_name", "SMTP"))

            # When a HTML is piped through --body flag, Content-Transfer-Encoding might change to ex: quoted-printable.
            # However when editing, we do not want to see quoted-printable garbage but plain text.
            # XX test should be added
            headers = [f"{k}: {v}" for k, v in e.as_message().items() if k.lower() != "content-transfer-encoding"]
            self.mail_file.write_text(PLAIN_DELIMITER.join((*headers, "", e.message())))

            # XDEPRECATED
            # So, we
            # 1) fetch the headers (except Content-Transfer-Encoding)
            # 2) insert blank line (between headers and the body)
            # 3) fetch the body while stripping out all <br> tags at the line ends
            #       (they should be re-inserted by envelope automatically and the readability increases)
            # XX test should be added
            # no_br_end = re.compile("<br\s?/?>$")
            # eml_text = [line for line in str(e)[:str(e).index("\n\n")].splitlines()
            #             if not line.startswith("Content-Transfer-Encoding")] +\
            #            [""] +\
            #            [no_br_end.sub("", line) for line in e.message().splitlines()]
            #
            # self.mail_file.write_text(PLAIN_DELIMITER.join(eml_text))

        self.file_assured = True
        return just_created

    def _make_envelope(self, attachment: Attachment):
        """ The format of the mail template is:
                Header: value
                Another-header: value
                Subject: value

                Body text begins after space.

            :raises RuntimeError if there is security concern about an attachment
        """
        try:
            self.text = self.mail_file.read_text()
        except FileNotFoundError:
            self.text = None

        if not self.text:  # user did not fill files in GUI
            return "Empty body text."

        if attachment and Config.get("jinja", "SMTP", get=bool):
            if self.apply_jinja(attachment) is False:
                return "Wrong jinja2 template."

        e = (Envelope
             .load(self.text)
             .signature("auto"))

        if not e.from_():
            e.from_(Config.get("email_from_name", "SMTP"))
        if not e.message() or not e.message().strip():
            return "Missing body text. Try writing 'Subject: text', followed by a newline and a text."
        if not e.subject():
            return "Missing subject. Try writing 'Subject: text', followed by a newline."
        if attachment:
            # set recipients
            e.to(attachment.mail)
            if attachment.cc:
                e.cc(attachment.cc)

            # attach the split CSV file
            if attachment.path and not attachment.used_in_body and Config.get('attach_files', 'SMTP', get=bool):
                e.attach(attachment.path, "text/csv", attachment.parser.attachment_name)

            # attach the paths from the path column (images, ...) in the split CSV file
            # If there is a trouble with an attachment,
            if Config.get('attach_paths_from_path_column', 'SMTP', get=bool):
                for path in attachment.get_paths_from_path_column():
                    try:
                        e.attach(path)
                    except FileNotFoundError as e:
                        return f"Cannot find the attachment: {e}"


        if Config.is_testing():
            e.recipients(clear=True)
            intended_to = attachment.mail if attachment else None
            e.to(Config.get('testing_mail'))
            # XX envelope might have method to delete message, like
            #  .message(False), .message(""), .message(text, alternative="replace")
            m = e.message()
            e.message("", alternative="auto") \
                .message("", alternative="plain") \
                .message("", alternative="html") \
                .message(f"This is testing mail only from Convey."
                         f" Don't be afraid, it was not delivered to: {intended_to}\r\n{m}")

        return e

    def get_envelope(self, attachment: Attachment = None) -> Envelope:
        """ :raises RuntimeError if there is security concern about an attachment """
        while True:
            just_created = self._assure_file()
            e = self._make_envelope(attachment)
            if isinstance(e, Envelope):
                if not just_created:  # even if Envelope is well generated, we invoke editing
                    return e
                e = "Editing just created file"
            # user fills GUI file, saves it and we get back to the method
            print(e)
            self.edit_text()

    def apply_jinja(self, attachment: Attachment):

        def print_attachment():
            """ Prints the attachment contents and prevent it to be attached.
                # XX may have header=False parameter to skip header.
            """
            attachment.used_in_body = True
            return attachment.path.read_text()

        def amount(count=2):
            """ Check if the attachment has at least count number of lines.
                Header is not counted.
            """
            if self.parser.settings["header"]:
                count += 1  # we skip header, so we must increment the count
            with attachment.path.open() as f:
                for i, _ in enumerate(f):
                    if i + 1 == count:
                        return True
            return False

        def row():
            """ Generate attachment row by row. Header is skipped. """
            with attachment.path.open() as f:
                if self.parser.settings["header"]:
                    next(f)
                for line in f:
                    yield line.strip().split(self.parser.settings["dialect"].delimiter)

        def joined(column: int, delimiter=", "):
            """ Return a column joined by delimiter  """
            return delimiter.join(r[column] for r in row())

        # Access first line fields
        first_row = next(row())
        try:
            self.text = Template(self.text).render(first_row=first_row,
                                                   row=row,
                                                   joined=joined,
                                                   amount=amount,
                                                   attachment=print_attachment)
        except exceptions.TemplateError as e:
            print(f"Template error: {e}")
            return False



from __future__ import annotations
from bdb import BdbQuit
from math import inf
from typing import TYPE_CHECKING, List, Optional, Union

from .config import Config
from .action import MergeAction
from .types import Type, Types
from .whois import Quota, UnknownValue

if TYPE_CHECKING:
    from .parser import Parser

Cell = Union[str, List[str]]
"""Even though standard fields in the CSV are of type str, while computing new columns,
    these might generate a list. Such list is immediately split to multiple lines while processing,
    but not when previewing the CSV table before the processing starts.
    """


class Field:
    def __init__(self, name, is_chosen=True, source_field: Optional[Field] = None, source_type=None, new_custom=None,
                 merged_from: Optional[Field] = None, merge_operation: Optional[MergeAction] = None):
        self.col_i = None
        "index of the field in parser.fields"
        self.col_i_original = None
        "original index before any sorting"
        self.parser: Optional[Parser] = None
        self.name = str(name)
        self.is_chosen = is_chosen
        self.is_selected = False
        self.merged_from = merged_from
        self.merge_operation = merge_operation
        "This field is merged from a remote file"
        self.possible_types = {}
        if isinstance(name, Type):
            self.type = name
        else:
            self.type = None
        self.is_new = False
        if source_field:
            self.is_new = True
            self.source_field = source_field
            self.source_type = source_type if type(source_type) is Type else getattr(Types, source_type)
            self.new_custom = new_custom
        else:
            self.source_field = self.source_type = self.new_custom = None

    def __repr__(self):
        return f"<Field {self.name}({self.type})>"

    def set_parser(self, parser):
        self.parser = parser

    def move(self, direction=0):
        """ direction = +1 right, -1 left """
        p = self.parser
        i, i2 = self.col_i, (self.col_i + direction) % len(p.fields)
        p.fields[i2].col_i, self.col_i = i, i2

        def swap(i, i2):
            """ Swap given iterable elements on the positions i and i2 """

            def _(it):
                it[i], it[i2] = it[i2], it[i]

            return _

        transposed = list(zip(*p.sample_parsed))
        list(map(swap(i, i2), [p.fields, transposed]))
        p.sample_parsed = list(map(list, zip(*transposed)))

    def toggle_chosen(self):
        self.is_chosen = not self.is_chosen

    @property
    def type(self):
        if self._type:
            return self._type
        if self.possible_types:
            return next(iter(self.possible_types))

    @type.setter
    def type(self, val):
        self._type = val
        if val:
            # for all not newly-added fields,
            # self.possible_types are re-build in identifier.identify_fields
            self.possible_types[val] = 100

    def color(self, v, shorten=False, line_chosen=True):
        """ Colorize single line of a value. Strikes it if field is not chosen. """
        v = str(v)  # ex: Types.http_status returns int
        if shorten:
            v = v[:17] + "..." if len(v) > 20 else v
        l = []
        if not self.is_chosen or not line_chosen:
            l.append("9")  # strike
        if self.is_selected:
            l.append("7")  # bold
        if self.is_new:
            l.append("33")  # yellow
        elif self.type is None or self.type == Types.plaintext:
            l.append("36")  # blue
        else:
            l.append("32")  # green
        s = "\033[" + ";".join(l) + f"m{v}\033[0m"
        return s

    def get(self, long=False, color=True, line_chosen=True):
        s = ""
        if long:
            if self.is_new:
                s = f"{self.name} from:\n{self.source_field}"
            elif self.has_clear_type():
                s = f"{self.name}\n   ({self.type})"
        if not s:
            s = self.name
        if color:
            s = "\n".join((self.color(c, line_chosen=line_chosen) for c in s.split("\n")))
        return s

    def has_clear_type(self):
        return self.type is not None and self.type != Types.plaintext

    def get_methods(self, target=None, start=None):
        if start is None:
            start = self.source_type
        if target is None:
            target = self.type
        return self.parser.identifier.get_methods_from(target, start, self.new_custom)

    def __str__(self):
        return self.name

    def get_samples(self, max_samples=inf, supposed_type=None, target_type=None):
        """ get few sample values of a field """
        c = min(len(self.parser.sample_parsed), max_samples)
        try:
            res = [self.parser.sample_parsed[line][self.col_i] for line in
                   range(0, c)]
        except IndexError:
            rows = []
            for l in self.parser.sample_parsed[slice(None, c)]:
                rows.append(self.compute_preview(l))
            res = rows
        if supposed_type and supposed_type.is_plaintext_derivable:
            rows, res = res.copy(), []
            for c in rows:
                for m in self.get_methods(Types.bytes if target_type == Types.charset else Types.plaintext, self.type):
                    c = m(c)
                res.append(c)
        return res

    def compute_preview(self, source_line: List[Cell]) -> Cell:
        """ source_line is registered under parser.sample_parsed """
        if Config.get("compute_preview") and self.source_field:
            try:
                c = source_line[self.source_field.col_i]
            except IndexError:
                # Needed for this complicated case:
                # convey "http://example.com" --web --field code,"x" --field text,code,url --field reg_m,text
                return "NOT COMPUTED YET"
            if c is None:
                # source column has not yet been resolved because of column resorting
                # (note this will not pose a problem when processing)
                return "..."
            # noinspection PyBroadException
            try:
                for l in self.get_methods():
                    if isinstance(c, list):
                        # resolve all items, while flattening any list encountered
                        c = [y for x in (l(v) for v in c) for y in (x if type(x) is list else [x])]
                    else:
                        c = l(c)
            except Quota.QuotaExceeded:
                c = "QUOTA EXCEEDED"
            except UnknownValue:
                c = "UNKNOWN"
            except BdbQuit:
                raise
            except Exception:
                c = "INVALID"
        elif Config.get("compute_preview") and self.merged_from:
            # this field is merged
            try:
                key = source_line[self.merge_operation.local_column.col_i]
                remote_lines: Optional[List[str]] = self.merge_operation.rows.get(key)
                if remote_lines:
                     # list only first line; user will not know that the line might be duplicated
                     # if key exist multiple times
                    c = remote_lines[0][self.merged_from.col_i]
                else:
                    c = "EMPTY"
            except IndexError:
                # Needed for this complicated case:
                # convey "http://example.com" --web --field code,"x" --field text,code,url --field reg_m,text
                return "NOT COMPUTED YET"
        else: # we do not want the preview to be computed
            c = "..."

        # add a newly computed value to source_parsed
        for _ in range(self.col_i - len(source_line) + 1):  # source_line is shorter than we need - fill the missing cols with Nones
            source_line.append(None)
        if type(c) is list:
            if len(c) == 1:
                # XX add comment when does this happen
                # Can we receive multiple values? Should not we join them?
                c = c[0]
            elif len(c) == 0:
                c = "NOTHING"
        if c is None:
            c = "NOTHING"
        source_line[self.col_i] = c
        return c



# Note: as this file is loaded before config.py initializing logging, we cannot use logger here
from abc import ABC, abstractmethod
from typing import Callable


class PickBase(ABC):
    default = None
    subtype: Callable

    @abstractmethod
    def get_lambda(self): pass

    def get_type_description(self):
        return self.subtype.__doc__.strip()


class PickMethod(PickBase):
    """ If you need to ask a question before computing values,
    make a class with methods that will be considered as options.
    User will be asked what computing option they tend to use,
    whilst the description is taken directly from the methods' __doc__ string.
    Than decorate with @PickMethod with optional default:str parameter that points to the default method.

     Ex: A user should be asked if a generator should return every value or perform a filtering.

     @PickMethod("all")
     class MyValues():
        ''' description shown in help '''

        def all(x):
            ''' return every value (this very text will be displayed to the user) '''
            return x

        def filtered(cls, x):
            ''' return only specific values '''
            if x in my_set:
                return x
    """

    def get_lambda(self, custom=None):
        if custom is None:
            custom = self.default
        for name in self._get_options():
            if name == custom:
                return getattr(self.subtype, name)
        else:
            raise NotImplementedError(f"Option {custom} has not been implemented for {self.subtype}, only {self._get_options()}.")

    def get_options(self):
        """ Return generator of options name and description tuples """
        return ((name, getattr(self.subtype, name).__doc__.strip()) for name in self._get_options())

    def _get_options(self):
        return (name for name in self.subtype.__dict__ if not name.startswith("_"))

    def __init__(self, default: str = None):
        self.default = default

    def __call__(self, subtype):
        self.subtype = subtype
        return self


class PickInput(PickBase):
    """ If your external function need to be setup with a variable first,
     decorate with @PickInput and register a function having two parameters. The second may have a default value.

    In this example, we let the user decide what should be the value of `format` before processing.
    All values will be formatted with the same pattern.

    @PickInput
    def time_format(val, format="%H:%i"):
        ''' this text will be displayed to the user '''
        return dateutil.parser.parse(val).strftime(format)

    """

    def get_lambda(self, custom=None):
        if custom:
            return lambda x: self.subtype(x, custom)
        else:
            return lambda x: self.subtype(x)

    def __init__(self, subtype):
        # since it takes about 5 ms and this is loaded directly from __init__.py, we postpone the loading here
        # another solution would be to implement a lazy loading from __init__.py:all
        import inspect
        self.subtype = subtype
        par = list(inspect.signature(subtype).parameters)
        if len(par) != 2:
            raise RuntimeError(f"Cannot import {subtype.__name__}, it has not got two parameters.")
        p = inspect.signature(subtype).parameters[par[1]]
        self.default = None if p.default is p.empty else p.default
        self.description = subtype.__doc__ or (f"Input {subtype.__name__} variable " + par[1])
        self.parameter_name = par[1]



# Env config file connection
import configparser
import glob
import logging
import os
import re
import sys
import webbrowser
from pathlib import Path
from shutil import copy
from subprocess import Popen, PIPE, call, run
from sys import exit
from time import sleep
from urllib.parse import quote

from appdirs import user_config_dir

# setup logging
# This cannot be in __init__.py so that we cannot reliably use logger in __init__.py, __main__.py and decorators.py.
# The reason is __init__.py gets launched when generic Python autocompletion searches for includable modules.
# If user would hit tab when writing `python3 -m [LETTER].[TAB]` into terminal, empty convey.log would have been created in the dir.
handlers = []
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter('%(message)s'))
console_handler.setLevel(logging.INFO)
handlers.append(console_handler)
try:
    file_handler = logging.FileHandler("convey.log")
    file_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    file_handler.setLevel(logging.WARNING)
    handlers.append(file_handler)
except PermissionError:
    file_handler = None
    print("Cannot create convey.log here at " + str(os.path.abspath(".")) + " – change directory please.")
    exit()
except FileNotFoundError:  # FileNotFoundError emitted when we are in a directory whose inode exists no more
    print("Current working directory doesn't exist.")
    exit()
logging.basicConfig(level=logging.INFO, handlers=handlers)

# mute noisy module (prints warning every time a validation fails which fails permanently when determining data type)
logging.getLogger('validate_email').setLevel(logging.ERROR)
logging.getLogger('filelock').setLevel(logging.WARNING)

# class init
logger = logging.getLogger(__name__)
Path.lexists = lambda self: self.is_symlink() or self.exists()  # not yet exist https://bugs.python.org/issue34137
default_path = Path(Path(__file__).resolve().parent, "defaults")  # path to the 'defaults' folder with templates

config_dir = user_config_dir("convey")
BOOLEAN_STATES = configparser.RawConfigParser.BOOLEAN_STATES
console_handler, file_handler, *_ = *logging.root.handlers, None


def get_path(file):
    """ Assures the file is ready, or creates a new one from a default. """
    global config_dir
    exists = True
    try:
        b = Path(file).lexists()
    except PermissionError as e:
        b = False  # ex we run program from /root under another user
    finally:
        if b:
            # check the .ini file is at current location
            file = Path(Path.cwd(), file)
        elif Path(Path(sys.argv[0]).parent, file).lexists():
            # file at program folder (I.E. at downloaded github folder)
            file = Path(Path(sys.argv[0]).parent, file)
        elif Path(config_dir, file).lexists():
            # INIT file at user config folder
            file = Path(config_dir, file)
        else:
            exists = False

    if exists:
        while not Path(file).exists():
            i = input(f"File on the path {file} may be a broken symlink. "
                      f"Mount it and press any key / 'q' for program exit / 'c' for recreating files / 'i' temporarily ignore: ")
            if i == "q":
                print("Exiting.")
                exit()
            elif i == "c":
                exists = False
                break
            elif i == 'i':
                return file

    if not exists or not Path(file).exists():
        # create INI file at user config folder or at program directory
        program_path = Path(sys.argv[0]).parent.resolve()
        # this is before Config file is load, we have to parse sys.argv directly
        yes = any(x in sys.argv for x in ['-y', '--yes', '-H', '--headless'])
        if not yes:
            yes = input(f"It seems this is a first run, since file {file} haven't been found."
                        f"\nShould we create default config files at user config folder ({config_dir})?"
                        f" Otherwise, they'll be created at program folder: {program_path} [Y/n] ") in ["", "Y", "y"]
        if yes:
            Path(config_dir).mkdir(parents=True, exist_ok=True)
        else:
            config_dir = program_path
        try:
            for filename in glob.glob(str(Path(default_path, '*.*'))):
                # do not overwrite existing files
                #  – user just might have deleted EML templates and they would not be happy to see their config.ini wiped
                if not Path(config_dir, Path(filename).name).exists():
                    copy(str(filename), config_dir)
            file = Path(f"{config_dir}/{file}")
        except Exception as e:
            print(e)
            print("Error creating program file {}. Exiting.".format(file))
            exit()

    return file


class Config:
    path = get_path("config.ini")
    cache = {}

    config = configparser.ConfigParser()
    config.read(path)

    # muted = False  # silence info text

    QUEUED_NAME = ".queues_lines.tmp"
    INVALID_NAME = ".invalidlines.tmp"
    UNKNOWN_NAME = "unknown"
    ABROAD_MARK = "@@"
    PROJECT_SITE = "https://github.com/CZ-NIC/convey/"
    verbosity: int = logging.INFO  # standard python3 logging level int

    @staticmethod
    def missing_dependency(library):
        logger.warning(f"Install {library} by `sudo apt install {library}` first or disable `--{library}` in config")
        input("Press any key...")
        raise KeyboardInterrupt

    @staticmethod
    def integrity_check():
        from .dialogue import is_yes  # since Config should be the very first package file to be loaded, we postpone this import
        # Config file integrity check (we may upgrade Convey from but some new parameters needs to be added manually)
        default_config = configparser.ConfigParser()
        default_ini = Path(default_path, "config.ini")
        default_config.read(default_ini)
        passed = True
        section_missing = []
        section_superfluous = []
        key_missing = {}  # [key] => section
        key_superfluous = []
        for section in default_config:
            missing_section = False
            if section not in Config.config:
                print(f"Missing section: {section}")
                section_missing.append(section)
                passed = False
                missing_section = True
                # continue
            for key in default_config[section]:
                if missing_section or key not in Config.config[section]:
                    print(f"Missing key {key} (defaulting to {repr(default_config[section][key])}) in section: {section}")
                    key_missing[key] = section
                    passed = False
        for section in Config.config:
            if section not in default_config:
                print(f"(Config has an unused section: {section})")
                section_superfluous.append(section)
                continue
            for key in Config.config[section]:
                if key not in default_config[section]:
                    print(f"Config has an unused key {key} in section: {section}")
                    key_superfluous.append(key)
        if not passed:
            # analyze current config file and get its section start line numbers
            config_lines = Config.path.read_text().splitlines(keepends=True)
            section_lines = {}  # [section name] = section start line number
            for i, line in enumerate(config_lines):
                try:
                    section = re.match("\[([^]]*)\]", line).group(1)
                except AttributeError:
                    pass
                else:
                    section_lines[section] = i

            # analyze the default config file
            default_lines = default_ini.read_text().splitlines(keepends=True)

            def get_key(iterable, match):
                """
                :rtype: yields tuple of (key name, slice start-stop line of the iterable)
                """
                comment = None
                anchors = tuple(iterable)
                for i, line in enumerate(iterable):  # add missing keys and sections
                    if line.startswith("#"):
                        if not comment:
                            comment = i
                        continue
                    if line.startswith(anchors):
                        key = re.match("[a-zA-Z0-9_]*", line).group(0)
                        if key in match:  # we have found one of the matching keys
                            yield key, slice(comment or i, i + 1)
                    comment = None

            # remove unused keys
            for key, slice_ in get_key(config_lines, key_superfluous):
                for i in range(slice_.start, slice_.stop):
                    config_lines[i] = ""
            # insert missing keys and sections
            for key, slice_ in get_key(default_lines, key_missing):
                block = default_lines[slice_]
                try:
                    start_line = section_lines[key_missing[key]]
                except KeyError:  # insert missing section
                    section_lines[key_missing[key]] = start_line = len(config_lines)
                    config_lines.append(f"\n[{key_missing[key]}]")
                config_lines[start_line] += "\n" + "".join(block)  # insert missing key
            if is_yes("Should I add the missing keys (and remove the unused) from your config file?"):
                Config.path.write_text("".join(config_lines))
                print("Config file has been successfully modified. Restart Convey.")
            else:
                print("Please write missing items into the config file before continuing.\nOpening", Config.path, "...")
                p = Popen(["xdg-open", Config.path], shell=False)
            exit()

    @staticmethod
    def init_verbosity(yes=False, verbosity=None, daemon=None):
        # Set up logging and verbosity
        if daemon:
            Config.set("daemon", True)
        if yes:
            Config.set("yes", True)
        if verbosity:
            Config.verbosity = verbosity
        else:
            # we reset the verbosity to the default value
            # usecase: client could ask the daemon for logging.DEBUG.
            # In the next request, they does not specify the verbosity. If not reset, the verbosity would still be logging.DEBUG.
            Config.verbosity = logging.INFO
        if Config.is_debug():
            if not verbosity:  # if user has not say the verbosity level, make it the most verbose
                Config.verbosity = logging.DEBUG
            if file_handler:
                file_handler.setLevel(logging.INFO)  # file handler to info level
            logging.getLogger("chardet").setLevel(logging.WARNING)  # we are not interested in chardet debug logs

        console_handler.setLevel(Config.verbosity)  # stream handler to debug level
        logging.getLogger().setLevel(min(Config.verbosity, logging.INFO))  # system sensitivity at least at INFO level

        logger.debug("Config file loaded from: {}".format(Config.path))

    @staticmethod
    def get_debugger():
        try:
            import ipdb as mod
        except ImportError:
            try:  # pudb not preffered as it wrote "NoneType: None" instead of exceptions
                import pudb as mod
            except ImportError:
                import pdb as mod
        return mod

    @staticmethod
    def error_caught(force=False):
        """ Return True if the user got the prompt (and possibly clean the error up then). """
        if Config.is_debug() or force:
            import traceback
            import sys

            mod = Config.get_debugger()
            type_, value, tb = sys.exc_info()
            traceback.print_exc()
            if tb:
                print("Post mortem")
                mod.post_mortem(tb)
            else:
                print("Lets debug. Hit n to get to the previous scope.")
                mod.set_trace()
            return True
        return False

    @staticmethod
    def is_debug():
        return Config.get('debug')

    @staticmethod
    def is_quiet():
        return Config.verbosity >= logging.WARNING

    @staticmethod
    def is_verbose():
        return Config.verbosity <= logging.DEBUG

    @staticmethod
    def is_testing():
        return Config.get('testing')

    @staticmethod
    def get(key, section='CONVEY', get=None):
        """

        :type get: type
                * if str, return value will always be str (ex: no conversion '1/true/on' to boolean happens)
                * if list, value tries to be split by a comma
                * if bool, it tries to convert to True/False from values 0/off/false 1/on/true
                * if int, you get integer or 0 on error (which means 0 for on/true)
                * if None
                    * bool for text 0/off/false 1/on/true
                    * None for text = '' or non-inserted value
                    * or any inserted value if non-conforming to previous possibilities
        """
        if key not in Config.cache:
            try:
                val = Config.config[section][key]
                if (get is None or get is bool) and val.lower() in BOOLEAN_STATES:
                    val = BOOLEAN_STATES[val.lower()]
                elif val == '':
                    val = get() if get else None
            except (configparser.Error, KeyError):
                return get() if get else None
            Config.cache[key] = val
        val = Config.cache[key]
        if get is str and type(val) is not str:
            try:
                return str(Config.config[section][key])
            except KeyError:
                return ''
        elif get is list:
            if val:
                return [x.strip() for x in val.split(",")]
            return []
        elif get is int:
            try:
                return int(val)
            except ValueError:
                return 0
        return val

    @staticmethod
    def github_issue(title, body):
        url = f"https://github.com/CZ-NIC/convey/issues/new?title={quote(title)}&body={quote(body)}"
        if len(url) > 2000:
            # strip ending url-encoded character (because things like "%2" instead of "%20" would break the URL)
            url = re.sub(r'%(\d{1,2})?$', '', url[:1987]) + "%7D%60%60%60"  # add newline and 3× backtick
        webbrowser.open(url)
        input(f"\nPlease submit a Github issue at {url}"
              "\nTrying to open issue tracker in a browser...")

    @staticmethod
    def set(key, val, section='CONVEY'):
        # XX was this needed? if val is None:
        #    Config.config.remove_option(section, key)
        # else:
        Config.cache[key] = val
        # XX Config.config.set(section, key, str(val))

    cache_dir = ""
    output = None  # True, False, None or str (path)

    @staticmethod
    def set_cache_dir(path):
        Config.cache_dir = path
        if not path.exists():
            path.mkdir()

    @staticmethod
    def get_cache_dir():
        """ Cache dir with ending slash. """
        return Config.cache_dir


def edit(path="config", mode=3, restart_when_done=False, blocking=False):
    """

    @param path: One of the keywords to edit default files in config folder or any pathlib.Path.
    @param mode: 1 text, 2 gui, 3 gui or text
    @param restart_when_done: If True, user is told to restart Convey when done.
    @type blocking: If True and GUI used, user is told to Hit Enter before continuing.
    """
    if type(mode) is str:  # from CLI
        mode = int(mode)
    d = {"template": Config.get("mail_template"), "template_abroad": Config.get("mail_template_abroad"),
         "uwsgi": "uwsgi.ini", "config": "config.ini"}
    if path in d:
        path = get_path(d[path])
    elif not isinstance(path, Path):
        input(f"Not found: {path}. Hit Enter...")
        return

    if restart_when_done:
        print(f"Opening {path}... restart Convey when done.")

    gui = True
    if mode & 2:
        # we cannot use xdg-open because template.eml would probably launch an e-mail client
        # app = Popen(['xdg-open', path], stdout=PIPE, stderr=PIPE)
        try:
            editor = run(["xdg-mime", "query", "default", "text/plain"], stdout=PIPE).stdout.split()[0]  # run: blocking, output
            app = Popen(["gtk-launch", editor, path], stdout=PIPE, stderr=PIPE)  # Popen: non blocking
        except FileNotFoundError:
            library = "xdg-utils libgtk-3-bin"
            input(f"Install {library} by `sudo apt install {library}`. Hit Enter to launch CLI editor.")
            gui = False
    elif mode & 1:
        gui = False

    if not gui:
        call(["editor", path])  # call: blocking, no output
    elif mode & 3 == 3:
        for _ in range(10):  # lets wait a second to be sure GUI app started
            sleep(0.1)
            p = app.poll()
            if p is not None:
                if p != 0:
                    # a GUI app have not started, let's launch a CLI terminal
                    call(["editor", path])
                    gui = False
                break

    if gui and blocking:
        input("Press Enter to continue...")


def get_terminal_size():
    try:
        # XX when piping the input IN, it writes
        # echo "434" | convey -f base64  --debug
        # stty: 'standard input': Inappropriate ioctl for device
        # I do not know how to suppress this warning.
        height, width = (int(s) for s in os.popen('stty size', 'r').read().split())
        return height, width
    except (OSError, ValueError):
        return 0, 0


# we are awaiting english replies in the subprocess application
subprocess_env = dict(os.environ)
subprocess_env["LANG"] = "en_US.UTF-8"



from __future__ import annotations  # remove as of Python3.11 in every file
import csv
import itertools
import logging
import re
from _csv import Error, reader
from copy import copy
from csv import Sniffer
from difflib import SequenceMatcher
from statistics import mean
from sys import exit
from typing import Tuple, List, TYPE_CHECKING

from .config import Config
from .decorators import PickBase
from .types import Types, graph, TypeGroup, Type, get_module_from_path

if TYPE_CHECKING:
    from .field import Field

logger = logging.getLogger(__name__)


class Identifier:

    def __init__(self, parser):
        self.parser = parser
        self.graph = None

    @staticmethod
    def get_methods_from(target, start, custom):
        """
        Returns the nested lambda list that'll receive a value from start field and should produce value in target field.
        :param target: field type name
        :param start: field type name
        :param custom: List of strings that are being used by TypeGroup.custom and PickBase
        :return: lambda[]
        """
        custom = copy(custom)

        def custom_code(e: str):
            def method(x):
                l = locals()
                try:
                    exec(compile(e, '', 'exec'), l)
                except Exception as exception:
                    code = "\n  ".join(e.split("\n"))
                    logger.error(f"Statement failed with {exception}.\n  x = '{x}'; {code}")
                    if not Config.error_caught():  # XX ipdb cannot quit with q here
                        input("We consider 'x' unchanged...")
                    return x
                x = l["x"]
                return x

            return method

        def regex(type_, search, replace=None):
            search = re.compile(search)

            def reg_m_method(s):
                match = search.search(s)
                if not match:
                    return ""
                groups = match.groups()
                if not replace:
                    if not groups:
                        return match.group(0)
                    return match.group(1)
                try:
                    return replace.format(match.group(0), *[g for g in groups])
                except IndexError:
                    logger.error(f"RegExp failed: `{replace}` cannot be used to replace `{s}` with `{search}`")
                    if not Config.error_caught():
                        input("We consider string unmatched...")
                    return ""

            def reg_s_method(s):
                match = search.search(s)
                if not match:
                    return ""
                if not replace:
                    return search.sub("", s)
                try:
                    # we convert "str{0}" → "\g<0>" (works better than conversion to a mere "\0" that may result to ambiguity
                    return search.sub(re.sub("{(\d+)}", r"\\g<\1>", replace), s)
                except re.error:
                    logger.error(f"RegExp failed: `{replace}` cannot be used to substitute `{s}` with `{search}`")
                    if not Config.error_caught():
                        input("We consider string unmatched...")
                    return ""

            return reg_s_method if type_ == Types.reg_s else reg_m_method

        path = graph.dijkstra(target, start=start)  # list of method-names to calculate new fields
        lambdas = []  # list of lambdas to calculate new field
        if not path:
            return []
        for i in range(len(path) - 1):
            lambda_ = Types.get_method(path[i], path[i + 1])
            if isinstance(lambda_, PickBase):
                # either the fields was added (has custom:List)
                # or is being computed in run_single_query() through get_computable_fields that makes us sure PickBase has a default
                lambda_ = lambda_.get_lambda(custom.pop(0) if custom is not None else None)
            elif not hasattr(lambda_, "__call__"):  # the field is invisible, see help text for Types; may be False, None or True
                continue
            lambdas.append(lambda_)

        if target.group == TypeGroup.custom:
            if target == Types.external:
                try:
                    lambda_ = getattr(get_module_from_path(custom[0]), custom[1])  # (module path, method name)
                    if isinstance(lambda_, PickBase):
                        lambda_ = lambda_.get_lambda(custom.pop(2) if len(custom) > 2 else None)
                    lambdas.append(lambda_)
                except IndexError:
                    raise ValueError(f"You must specify which method should be used in {custom[0]}")
                except AttributeError:
                    raise ValueError(f"Cannot find method {custom[1]} in {custom[0]}")
            elif target == Types.code:
                if type(custom) is list and len(custom) == 1:
                    custom = custom[0]  # code accepts a string
                lambdas += [custom_code(custom)]
            elif target in [Types.reg, Types.reg_m, Types.reg_s]:
                lambdas += [regex(target, *custom)]  # custom is in the form (search, replace)
            else:
                raise ValueError(f"Unknown type {target}")

        logger.debug(f"Preparing path from {start} to {target}: " + ", ".join([str(p) for p in path])
                     + " ('" + "', '".join(custom) + "')" if custom else "")
        return lambdas

    @staticmethod
    def get_sample(source_file):
        sample = []
        first_line = ""
        is_pandoc = False
        with open(source_file, 'r') as csv_file:
            for i, row in enumerate(csv_file):
                if i == 0:
                    first_line = row
                if i == 1 and not row.replace("-", "").replace(" ", "").strip():
                    is_pandoc = True
                    continue
                sample.append(row)
                if i == 8:  # sniffer needs 7+ lines to determine dialect, not only 3 (/mnt/csirt-rook/2015/06_08_Ramnit/zdroj), I dont know why
                    break
        return first_line.strip(), sample, is_pandoc
        # csvfile.seek(0)
        # csvfile.close()

    def guess_dialect(self, sample):
        sniffer = Sniffer()
        sample_text = "".join(sample)
        try:
            dialect = sniffer.sniff(sample_text)
            has_header = sniffer.has_header(sample_text)
            if re.match("[a-z]", dialect.delimiter.lower()):  # we do not allow letters to be delimiters, seems like non-sense
                raise Error
        except Error:  # delimiter failed – maybe there is an empty column: "89.187.1.81,06-05-2016,,CZ,botnet drone"
            if sample_text.strip() == "":
                print("The file seems empty")  # XX I got here once after a clean installation at 26.11.2019
                exit()

            # header detection
            l = [line.strip() for line in sample]
            if len(l[1:]) > 0:
                header_to_rows_similarity = mean([SequenceMatcher(None, l[0], it).ratio() for it in l[1:]])
                if len(l[1:]) > 1:
                    rows_similarity = mean([SequenceMatcher(None, *comb).ratio() for comb in itertools.combinations(l[1:], 2)])
                    has_header = rows_similarity > header_to_rows_similarity + 0.1  # it seems that first line differs -> header
                else:
                    has_header = header_to_rows_similarity < 0.5
            else:
                has_header = False

            try:
                s = sample[1]  # we do not take header (there is no empty column for sure)
            except IndexError:  # there is a single line in the file
                s = sample[0]
            delimiter = ""
            for dl in (",", ";", "|"):  # lets suppose the doubled sign is delimiter
                if s.find(dl + dl) > -1:
                    delimiter = dl
                    break
            if not delimiter:  # try find anything that resembles to a delimiter
                for dl in (",", ";", "|"):
                    if s.find(dl) > -1:
                        delimiter = dl
                        break
                else:
                    if self.parser.is_pandoc and s.count(" ") > 1:
                        delimiter = " "
            dialect = csv.unix_dialect
            if delimiter:
                dialect.delimiter = delimiter
        if not dialect.escapechar:
            dialect.escapechar = '\\'
        # dialect.quoting = 3
        dialect.doublequote = True

        seems_single = False
        if len(sample) == 1:
            # there is single line in sample = in the input, so this is definitely not a header
            has_header = False
            if dialect.delimiter not in [".", ",", "\t"] and "|" not in sample_text:
                # usecase: short one-line like "convey hello" would produce stupid "l" delimiter
                # XX should be None maybe, let's think a whole row is a single column – but then we could not add columns
                dialect.delimiter = "|"
                seems_single = True
        if dialect.delimiter == "." and "," not in sample_text:
            # let's propose common use case (bare list of IP addresses) over a strange use case with "." delimiting
            dialect.delimiter = ","
        return dialect, has_header, seems_single

    def identify_fields(self, quiet=False):
        """
        Identify self.parser.fields got in __init__
        Sets them possible types (sorted, higher score mean bigger probability that the field is of that type)
        :type quiet: bool If True, we do not raise exception when sample cannot be processed.
                            Ex: We attempt consider user input "1,2,3" as single field which is not, we silently return False
        """
        samples = [[] for _ in self.parser.fields]
        if len(self.parser.sample) == 1:  # we have too few values, we have to use them
            s = self.parser.sample[:1]
        else:  # we have many values and the first one could be header, let's omit it
            s = self.parser.sample[1:]

        for row in reader(s, skipinitialspace=self.parser.is_pandoc, dialect=self.parser.dialect) if self.parser.dialect else [s]:
            for i, val in enumerate(row):
                try:
                    samples[i].append(val)
                except IndexError:
                    if not quiet:
                        print("It seems rows have different lengths. Cannot help you with column identifying.")
                        print("Fields row: " + str([(i, str(f)) for i, f in enumerate(self.parser.fields)]))
                        print("Current row: " + str(list(enumerate(row))))
                        if not Config.error_caught():
                            input("\n... Press any key to continue.")
                    return False

        for i, field in enumerate(self.parser.fields):
            possible_types = {}
            for type_ in Types.get_guessable_types():
                score = type_.check_conformity(samples[i], self.parser.has_header, field)
                if score:
                    possible_types[type_] = score
            # Use with Python3.8
            # possible_types = {type_: score for type_ in Types.get_guessable_types()
            #                   if (score := type_.check_conformity(samples[i], self.parser.has_header, field))}
            if field.type:
                possible_types[field.type] = 100

            if possible_types:  # sort by biggest score - biggest probability the column is of this type
                field.possible_types = {k: v for k, v in sorted(possible_types.items(), key=lambda k: k[1], reverse=True)}
            logger.debug(f"Possible type of the field '{field}': {possible_types}")
        return True

    def get_fitting_type(self, source_field_i, target_field, try_plaintext=False):
        """ Loops all types the field could be and return the type best suited method for compute new field. """
        _min = 999
        fitting_type = None
        possible_fields = list(self.parser.fields[source_field_i].possible_types)
        dijkstra = graph.dijkstra(target_field)  # get all fields that new_field is computable from
        for _type in possible_fields:
            # loop all the types the field could be, loop from the FieldType we think the source_col correspond the most
            # a column may have multiple types (url, hostname), use the best
            if _type not in dijkstra:
                continue
            i = dijkstra[_type]
            if i < _min:
                _min, fitting_type = i, _type
        if not fitting_type and try_plaintext and Types.plaintext in dijkstra:
            # try plaintext field as the last one. Do not try it earlier, usecase:
            # we want to produce reg_s from base64. If we inserted plaintext earlier,
            #  fitting_type would be plaintext since it is a step nearer - but the field would not be decoded
            return Types.plaintext
        return fitting_type

    def get_fitting_source_i(self, target_type: Type, try_hard=False) -> List[int]:
        """ Get list of source_i that may be of such a field type that new_field would be computed effectively.
            Note there is no fitting column for TypeGroup.custom, if you try_hard, you receive first column as a plaintext.

            Sorted by relevance.
        """
        possible_cols = {}
        if target_type.group != TypeGroup.custom:
            valid_types = graph.dijkstra(target_type)
            for val in valid_types:  # loop from the best suited type
                for i, f in enumerate(self.parser.fields):  # loop from the column we are most sure with its field type
                    if val in f.possible_types:
                        possible_cols[i] = f.possible_types[val]
                        break
        if not possible_cols and try_hard and target_type.is_plaintext_derivable:
            # because any plaintext would do (and no plaintext-only type has been found), take the first column
            possible_cols = [0]
        return list(possible_cols)

    def get_fitting_source(self, target_type: Type, *task) -> Tuple[Field, Type, List]:
        """
        For a new field, we need source column and its field type to compute new field from.
        :rtype: source_field: Field, source_type: Type, custom: List[str]
        :param target_type: Type
        :type task: List[str]: [COLUMN],[SOURCE_TYPE],[CUSTOM],[CUSTOM...]
            COLUMN: int|existing name
            SOURCE_TYPE: field type name|field type usual names
            CUSTOM: any parameter
        """
        source_col_i = None
        source_type = None
        source_col_candidates: List[Field] = []
        task = list(task)

        if Config.is_debug():
            print(f"Getting type {target_type} with args {task}")

        # determining source_col_i from a column candidate
        column_candidate = task.pop(0) if len(task) else None
        if column_candidate:  # determine COLUMN
            source_col_i = self.get_column_i(column_candidate)  # get field by exact column name, ID or type
            if source_col_i is None:
                if len(task) and target_type.group != TypeGroup.custom:
                    print(f"Invalid field type {task[0]}, already having defined by {column_candidate}")
                    exit()
                task.insert(0, column_candidate)  # this was not COLUMN but SOURCE_TYPE or CUSTOM, COLUMN remains empty
        if source_col_i is None:  # get a column whose field could be fitting for that target_tape or any column as a plaintext
            try:
                source_col_candidates = [self.parser.fields[i] for i in self.get_fitting_source_i(target_type, True)]
                source_col_i = source_col_candidates[0].col_i
            except IndexError:
                pass

        # determining source_type
        source_type_candidate = task.pop(0) if len(task) else None
        if source_type_candidate:  # determine SOURCE_TYPE
            source_type = Types.find_type(source_type_candidate)
            if source_type:
                # We have to choose the right column - if there are some source_col_candidates it means the column
                # was not determined exactly, we are just guessing. Get the one of the candidates whose type is nearest
                # to the demanded source_type.
                # Usecase:
                #   * command: `--field incident_contact,source_ip`
                #   * columns: we have only `hostname` (or only `any_ip`) amongst columns.
                #   * This will check there is no source_ip amongst columns
                #       and instead of letting it be `source_type = source_ip` which would fail when resolving hostname:
                #   * either corrects `source_type = hostname` (source_ip -> ip -> hostname)
                #   * or corrects `source_type = any_ip` (any_ip -> source_ip)

                # XX as of Python3.8, replace the next statement with this
                # try:
                #     best_candidate = min((len(path), field.col_i, field.type) for field in source_col_candidates
                #                             if (path:=(graph.dijkstra(field.type, start=source_type) \
                #                             or graph.dijkstra(source_type, start=field.type)) is not False))
                #     source_type = best_candidate[2]
                # except ValueError:
                #     pass

                # XX we might rather have a method: Type("source_ip").get_converted() returning Type("ip")
                source_type = Type("ip") if source_type in (Type("source_ip"), Type("destination")) else source_type
                best_candidate = None
                for field in source_col_candidates:
                    path = graph.dijkstra(field.type, start=source_type)\
                            or graph.dijkstra(source_type, start=field.type)
                    if path:
                        best_candidate = min((len(path), field.col_i, field.type), best_candidate or (float('INF'),))
                        source_type = best_candidate[2]

            elif target_type.group == TypeGroup.custom:
                # this was not SOURCE_TYPE but CUSTOM, for custom fields, SOURCE_TYPE may be implicitly plaintext
                #   (if preprocessing ex: from base64 to plaintext is not needed)
                task.insert(0, source_type_candidate)
                # source_type = Types.plaintext
            else:
                print(f"Cannot determine new field from {source_type_candidate}")
                exit()

        # determining missing info
        if source_col_i is not None and not source_type:
            try:
                source_type = self.get_fitting_type(source_col_i, target_type, try_plaintext=True)
            except IndexError:
                print(f"Column ID {source_col_i + 1} does not exist. We have these so far: " +
                      ", ".join([f.name for f in self.parser.fields]))
                exit()
            if not source_type:
                print(f"We could not identify a method how to make '{target_type}' from '{self.parser.fields[source_col_i]}'")
                exit()
        if source_type and source_col_i is None:
            # searching for a fitting type amongst existing columns
            # [source col i] = score (bigger is better)
            possibles = {i: t.possible_types[source_type] for i, t in enumerate(self.parser.fields)
                         if source_type in t.possible_types}
            try:
                source_col_i = sorted(possibles, key=possibles.get, reverse=True)[0]
            except IndexError:
                f = [f for f in self.parser.fields if not f.is_new]
                if len(f) == 1:
                    # there is just a single non-computed (and thus unidentified) column
                    # since we are forcing source_type, let's pretend the column is of this type
                    source_col_i = f[0].col_i
                else:
                    print(f"No suitable column of type '{source_type}' found to make field '{target_type}'")
                    exit()

        if not source_type or source_col_i is None:
            print(f"No suitable column found for field '{target_type}'")
            exit()

        try:
            f = self.parser.fields[source_col_i]
        except IndexError:
            print(f"Column ID {source_col_i + 1} does not exist, only these: " + ", ".join(f.name for f in self.parser.fields))
            exit()

        # Check there is a path between nodes and that path is resolvable
        path = graph.dijkstra(target_type, start=source_type)
        if path is False:
            print(f"No suitable path from '{f.name}' treated as '{source_type}' to '{target_type}'")
            exit()
        for i in range(len(path) - 1):  # assure there is a valid method
            try:
                Types.get_method(path[i], path[i + 1])
            except LookupError:
                print(f"Path from '{f.name}' treated as '{source_type}' to '{target_type}' blocked at {path[i]} – {path[i + 1]}")

        if Config.is_debug():
            print(f"Preparing type {target_type} of field={f}, source_type={source_type}, custom={task}, path={path}")
        return f, source_type, task

    def get_column_i(self, column, check=False):
        """
        Useful for parsing user input COLUMN from the CLI args.
        :type column: object Either column ID (ex "1" points to column index 0), position from the right (-1, ...),
                        desired type or an exact column name or the field
        :type check: If not False and not found, error is raised and quit. If str, this string will be included in the error text.
        :rtype: int Either column_i or None if not found.
        """
        source_col_i = None
        if hasattr(column, "col_i"):
            return column.col_i
        if re.match(r"-?\d+$", column):  # number of column
            i = int(column)
            source_col_i = i - 1 if i > 0 else i  # ID=1 -> col_i=0 (first column), ID=-1 -> col_i=-1 (last added)
        elif column in self.parser.first_line_fields:  # exact column name
            source_col_i = self.parser.first_line_fields.index(column)
        else:  # as of Python3.8, shorten to elif searched_type :=...
            searched_type = Types.find_type(column)  # get field by its type
            if searched_type:
                source_col_i = (next((f.col_i for f in self.parser.fields if f.type == searched_type), None) or
                                next((f.col_i for f in self.parser.fields if searched_type in f.possible_types), None))
        if check and (source_col_i is None or len(self.parser.fields) <= source_col_i):
            logger.error(f"Cannot identify COLUMN {column}" + (" " + check if type(check) is str else "") +
                         ", put there an exact column name, its type, the numerical order starting with 1, or with -1.")
            exit()
        return source_col_i
